/*********************************************************************/
/* tclparse V2.00 from 14.March 1997                                 */
/* Copyright 1996/97 Stephan Weiss, Heiko Grossmann                  */
/*                   University of Stuttgart, Germany                */
/* For details see the copyright file included in the distribution.  */
/* This notice must not be removed from the file.                    */
/*                                                                   */
/* routines for output                                               */
/*                                                                   */
/*********************************************************************/

#include "../include/const.h"
#include "../include/display.h"
#include "../include/file.h"
#include "../include/parsing.h"
#include <string.h>
#include <iostream.h>
#include <fstream.h>


CDisplay::CDisplay( void )
{
  barElement = new char[ 1 ];
  strcpy( barElement,"*" );

  lineElement = new char[ 1 ];
  strcpy( lineElement,"-" );

  separatorLength = 40;

  separator = new char[ separatorLength ];
  separator[0] = '\0';

  for( int i=1; i <= separatorLength ; i++ )
    strcat( separator, lineElement );
}


CDisplay::~CDisplay( void )
{
  delete lineElement;
  delete barElement;
  delete separator;
}


void CDisplay::Procedure( char *string )
{
  if( !flag.procedures )
    return;

  char message[ wordLength ];

  strcpy( message, "( procedure ) " );
  strcat( message, string );
  strcat( message, "..." );
  cout << message << '\n';
}


void CDisplay::ProcedureEnd( char *string )
{
  if( !flag.procedures )
    return;

  char message[ 100 ];

  strcpy( message, "( end of procedure ) " );
  strcat( message, string );
  strcat( message, "..." );
  cout << message << '\n';
}


void CDisplay::TclCommand( char *string )
{
  if( !flag.commands )
    return;

  char message[ 100 ];

  strcpy( message, "( tcl command ) " );
  strcat( message, string );
  strcat( message, "..." );
  cout << message << '\n';
}


void CDisplay::TkCommand( char *string )
{
  if( !flag.commands )
    return;

  char message[ 100 ];

  strcpy( message, "( tk command ) " );
  strcat( message, string );
  strcat( message, "..." );
  cout << message << '\n';
}


void CDisplay::Variable( char *string )
{
  if( !flag.variables )
    return;

  char message[ 100 ];

  strcpy( message, "( variable ) " );
  strcat( message, string );
  strcat( message, "..." );
  cout << message << '\n';
}


void CDisplay::Widget( char *string )
{
  if( !flag.widgets )
    return;

  char message[ 100 ];

  strcpy( message, "( widget ) " );
  strcat( message, string );
  strcat( message, "..." );
  cout << message << '\n';
}


void CDisplay::Error( char *message )
{
  if( !flag.errors )
    return;

  char procedureName[ wordLength ];
  char commandName[ wordLength ];
  char title[ wordLength ];
  char dummy[ wordLength ];

  //Zuerst den Namen der Prozedur ermitteln, in der der Fehler auftrat.
  strcpy( procedureName, actualFile->stack.Top( procedures ) );

  //Den Namen des Befehls ermitteln, in dem der Fehler auftrat.
  strcpy( commandName, actualFile->stack.Top( commands ) );

  //Die angegebene Meldung ausgeben.
  cout << "\nERROR -> " << message << "\n";

  //Informationen ueber die Position des Fehlers ausgeben.
  cout << "( ";

  if( strcmp( procedureName, "" ) )
    cout << "'" << procedureName << "' - ";

  if( strcmp( commandName, "" ) )
    cout << "'" << commandName << "' - ";

  cout << actualFile->error.y << " ):\n";

  //Einen Teil des Programmcodes ausgeben, in dem der Fehler auftrat.
  if( flag.code )
    ErrorProgramCode();
}


void CDisplay::Error( char *part1, char *part2 )
{
  if( !flag.errors )
    return;

  char message[ strlen( part1 ) + strlen( part2 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  Error( message );
}


void CDisplay::Error( char *part1 ,char *part2, char *part3 )
{
  if( !flag.errors )
    return;

  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  Error( message );
}


void CDisplay::Error( char *part1 ,char *part2, char *part3, char *part4 )
{
  if( !flag.errors )
    return;

  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + strlen( part4 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  strcat( message, part4 );
  Error( message );
}


void CDisplay::Error( char *part1 ,char *part2, char *part3, char *part4, char *part5 )
{
  if( !flag.errors )
    return;

  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + strlen( part4 ) + strlen( part4 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  strcat( message, part4 );
  strcat( message, part5 );
  Error( message );
}


void CDisplay::ErrorProgramCode( void )
{
  //Aktuelle Dateiposition speichern.
  streampos dummy = actualFile->Tell();

  //Dateizeiger auf den Anfang der Zeile drei Reihen vorher setzen.
  actualFile->Seek( actualFile->GetErrorOffset() );

  cout << separator << '\n';

  int lineNr = actualFile->cursor.y - 3;

  //Insgesamt sieben Zeilen einlesen und ausgeben.
  for( int i=0 ; i < 7 ; i++ )
    {
      actualFile->GetLine();

      if( actualFile->Eof() )
	break;

      cout << lineNr << ":" << Buffer() << '\n';
      lineNr++;
    }

  cout << separator << "\n\n";
  actualFile->Seek( dummy );
}


void CDisplay::InternalError( char *procName, char *code )
{
  cout << "internal error in procedure '" << procName << "'\n";
  cout << "caused by '" << code << "'\n\n";
}


void CDisplay::Message( char *message )
{
  cout << message << "\n\n";
}


void CDisplay::Message( char *part1, char *part2 )
{
  char message[ strlen( part1 ) + strlen( part2 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  display.Message( message );
}


void CDisplay::Message( char *part1, char *part2, char *part3 )
{
  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  display.Message( message );
}


void CDisplay::Message( char *part1, char *part2, char *part3, char *part4 )
{
  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + strlen( part4 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  strcat( message, part4 );
  display.Message( message );
}


void CDisplay::NewProcedure( char *name, char *args )
{
  cout << separator << '\n';
  cout << "procedure : " << name << '\n';
  cout << "arguments : " << args << '\n';
}


void CDisplay::NewVariable( char *name )
{
  cout << separator << '\n';
  cout << "variable  : " << name << '\n';
}


void CDisplay::Title( char *message )
{
  //Die Laenge der Meldung wird benoetigt, um die Laenge des Rahmens
  //berechnen zu koennen.
  int len = strlen( message );

  //Der Wert wird um 4 erhoeht, um auf jeder Seite ein Zeichen Platz
  //zwischen Rahmen und Meldung zu haben.
  len += 4;

  //Eine Variable fuer die Rahmenleiste erzeugen.
  char bar[ len ];
  strcpy( bar,"" );

  //Die Leiste des Rahmens erzeugen
  for( int index=0 ; index < len ; index++ )
    strcat( bar,barElement );

  //Die Ausgabe.
  cout << '\n';
  cout << bar << '\n';
  cout << barElement << ' ' << message << ' ' << barElement << '\n';
  cout << bar << "\n\n";
}


void CDisplay::Title( char *part1, char *part2 )
{
  char message[ strlen( part1 ) + strlen( part2 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  Title( message );
}


void CDisplay::Title( char *part1, char *part2, char *part3 )
{
  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  Title( message );
}


void CDisplay::Title( char *part1, char *part2, char *part3, char *part4 )
{
  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + strlen( part4 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  strcat( message, part4 );
  Title( message );
}


void CDisplay::Title( char *part1, char *part2, char *part3, char *part4, char *part5 )
{
  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + strlen( part4 ) + strlen( part5 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  strcat( message, part4 );
  strcat( message, part5 );
  Title( message );
}


void CDisplay::UnknownOption( char *option, char *command )
{
  //Handelt es sich bei der fehlerhaften Option um eine Variablensubstitution,
  //so wird nur eine Warnung ausgegeben.
  if( option[0] == '$' )
    Warning( "unknown option for command '", command, "': '", option, "'" );

  else
    Error( "unknown option for command '", command, "': '", option, "'" );
}


void CDisplay::UnknownVariable( char *varName )
{
  //Handelt es sich bei der fehlerhaften Option um eine Variablensubstitution,
  //so wird nur eine Warnung ausgegeben.
  if( varName[0] == '$' )
    Warning( "unknown variable: '", varName, "'" );

  else
    Error( "unknown variable: '", varName, "'" );
}


void CDisplay::Warning( char *message )
{
  if( !flag.warnings )
    return;

  char procedureName[ wordLength ];
  char commandName[ wordLength];
  char title[ wordLength ];
  char dummy[ wordLength ];

  //Zuerst den Namen der Prozedur ermitteln, in der der Fehler auftrat.
  strcpy( procedureName, actualFile->stack.Top( procedures ) );

  //Den Namen des Befehls ermitteln, in dem der Fehler auftrat.
  strcpy( commandName, actualFile->stack.Top( commands ) );

  //Die angegebene Meldung ausgeben.
  cout << "\nWARNING -> " << message << "\n";

  //Informationen ueber die Position des Fehlers ausgeben.
  cout << "( ";

  if( strcmp( procedureName, "" ) )
    cout << "'" << procedureName << "' - ";

  if( strcmp( commandName, "" ) )
    cout << "'" << commandName << "' - ";

  cout << actualFile->error.y << " ):\n";

  //Einen Teil des Programmcodes ausgeben, in dem der Fehler auftrat.
  if( flag.code )
    ErrorProgramCode();
}


void CDisplay::Warning( char *part1, char *part2 )
{
  if( !flag.warnings )
    return;

  char message[ strlen( part1 ) + strlen( part2 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  Warning( message );
}


void CDisplay::Warning( char *part1 ,char *part2, char *part3 )
{
  if( !flag.warnings )
    return;

  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  Warning( message );
}


void CDisplay::Warning( char *part1 ,char *part2, char *part3, char *part4 )
{
  if( !flag.warnings )
    return;

  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + strlen( part4 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  strcat( message, part4 );
  Warning( message );
}


void CDisplay::Warning( char *part1 ,char *part2, char *part3, char *part4, char *part5 )
{
  if( !flag.warnings )
    return;

  char message[ strlen( part1 ) + strlen( part2 ) + strlen( part3 ) + strlen( part4 ) + strlen( part5 ) + 1 ];
  strcpy( message, part1 );
  strcat( message, part2 );
  strcat( message, part3 );
  strcat( message, part4 );
  strcat( message, part5 );
  Warning( message );
}


CDisplay display;
