/*********************************************************************/
/* tclparse V2.00 from 14.March 1997                                 */
/* Copyright 1996/97 Stephan Weiss, Heiko Grossmann                  */
/*                   University of Stuttgart, Germany                */
/* For details see the copyright file included in the distribution.  */
/* This notice must not be removed from the file.                    */
/*                                                                   */
/* routines needed for parsing                                       */
/*                                                                   */
/*********************************************************************/

#include "../include/const.h"
#include "../include/file.h"
#include "../include/display.h"
#include "../include/parsing.h"
#include "../include/tclcmd.h"
#include "../include/tkcmd.h"
#include "../include/handler.h"
#include <string.h>
#include <stdlib.h>


//Viele Prozeduren benutzen Kopien ihrer Argumente. Falls der Puffer als Argument
//angegeben wurde, wird er so nicht veraendert. Umstaendlich, aber sicher.


void AddCommand( CCommand &command )
{
  char name[ strlen( command.Name() )+1 ];
  strcpy( name, Buffer() );

  char commandID[ strlen( command.CommandID() )+1 ];
  strcpy( commandID, Buffer() );

  char list[ strlen( command.List() )+1 ];
  strcpy( list, Buffer() );

  actualFile->handler.AddCommand( name, commandID, list, command.Min(), command.Max(), command.Calls() );
}


void AddCommand( char *name, char *commandID, char *list, int min=0, int max=0, int calls=0 )
{
  actualFile->handler.AddCommand( name, commandID, list, min, max, calls );
}


void AddImage( char *name, char *type, int calls=0 )
{
  if( name[0] != '$' )
    actualFile->handler.AddImage( name, type, calls );
}


void AddProcedure( char *name, char *arguments, int min=0, int max=0, int calls=0 )
{
  actualFile->handler.AddProcedure( name, arguments, min, max, calls );
}


int AddVariable( CVariable &variable )
{
  char name[ strlen( variable.Name() )+1 ];
  strcpy( name, Buffer() );

  char environment[ strlen( variable.Environment() )+1 ];
  strcpy( environment, Buffer() );

  return actualFile->handler.AddVariable( name, environment, variable.Status(), variable.Calls() );
}


int AddVariable( char *name, char *environment, int status=0, int calls=0 )
{
  if( name[0] != '$' )
    return actualFile->handler.AddVariable( name, environment, status, calls );

  else
    return 0;
}


void AddWidget( char *name, char *type, int calls=0 )
{
  actualFile->handler.AddWidget( name, type, calls );
}


char *Buffer( void )
{
  return actualFile->Buffer();
}


void CopyToBuffer( char *string )
{
  actualFile->CopyToBuffer( string );
}


//Die folgenden sechs Prozeduren kopieren den entsprechenden Inhalt in den
//Puffer und liefern einen Zeiger auf diesen zurueck.
char *Actual( dataType id )
{
  switch( id )
    {
    case atom:
      CopyToBuffer( actualFile->atom.actual );
      return Buffer();

    case symbol:
      CopyToBuffer( actualFile->symbol.actual );
      return Buffer();

    case word:
      CopyToBuffer( actualFile->word.actual );
      return Buffer();

    case unit:
      CopyToBuffer( actualFile->unit.actual );
      return Buffer();

    default:
      display.InternalError( "Actual", "unknown option" );
      CopyToBuffer( "\0" );
      return Buffer();
    }
}


char *ActualID( dataType id )
{
  switch( id )
    {
    case atom:
      CopyToBuffer( actualFile->atom.actualID );
      return Buffer();

    case symbol:
      CopyToBuffer( actualFile->symbol.actualID );
      return Buffer();

    case word:
      CopyToBuffer( actualFile->word.actualID );
      return Buffer();

    case unit:
      CopyToBuffer( actualFile->unit.actualID );
      return Buffer();

    default:
      display.InternalError( "ActualID", "unknown option" );
      CopyToBuffer( "\0" );
      return Buffer();
    }
}


char *Last( dataType id )
{
  switch( id )
    {
    case atom:
      CopyToBuffer( actualFile->atom.last );
      return Buffer();

    case symbol:
      CopyToBuffer( actualFile->symbol.last );
      return Buffer();

    case word:
      CopyToBuffer( actualFile->word.last );
      return Buffer();

    case unit:
      CopyToBuffer( actualFile->unit.last );
      return Buffer();

    default:
      display.InternalError( "Last", "unknown option" );
      CopyToBuffer( "\0" );
      return Buffer();
    }
}


char *LastID( dataType id )
{
  switch( id )
    {
    case atom:
      CopyToBuffer( actualFile->atom.lastID );
      return Buffer();

    case symbol:
      CopyToBuffer( actualFile->symbol.lastID );
      return Buffer();

    case word:
      CopyToBuffer( actualFile->word.lastID );
      return Buffer();

    case unit:
      CopyToBuffer( actualFile->unit.lastID );
      return Buffer();

    default:
      display.InternalError( "LastID", "unknown option" );
      CopyToBuffer( "\0" );
      return Buffer();
    }
}


char *Next( dataType id )
{
  char dummy[ bufferSize ];

  switch( id )
    {
    case atom:
      CopyToBuffer( actualFile->atom.next );
      return Buffer();

    case symbol:
      CopyToBuffer( actualFile->symbol.next );
      return Buffer();

    case word:
      actualFile->Save();
      actualFile->EnableScanFlag();
      GetWord();
      strcpy( dummy, Buffer() );
      actualFile->DisableScanFlag();
      actualFile->Restore();
      CopyToBuffer( dummy );
      return Buffer();

    case unit:
      actualFile->Save();
      actualFile->EnableScanFlag();
      GetUnit();
      strcpy( dummy, Buffer() );
      actualFile->DisableScanFlag();
      actualFile->Restore();
      CopyToBuffer( dummy );
      return Buffer();

    default:
      display.InternalError( "Next", "unknown option" );
      CopyToBuffer( "\0" );
      return Buffer();
    }
}


char *NextID( dataType id )
{
  char dummy[ bufferSize ];

  switch( id )
    {
    case atom:
      CopyToBuffer( actualFile->atom.nextID );
      return Buffer();

    case symbol:
      CopyToBuffer( actualFile->symbol.nextID );
      return Buffer();

    case word:
      actualFile->Save();
      actualFile->EnableScanFlag();
      GetWord();
      ActualID( word );
      strcpy( dummy, Buffer() );
      actualFile->DisableScanFlag();
      actualFile->Restore();
      CopyToBuffer( dummy );
      return Buffer();

    case unit:
      actualFile->Save();
      actualFile->EnableScanFlag();
      GetUnit();
      ActualID( unit );
      strcpy( dummy, Buffer() );
      actualFile->DisableScanFlag();
      actualFile->Restore();
      CopyToBuffer( dummy );
      return Buffer();

    default:
      display.InternalError( "NextID", "unknown option" );
      return Buffer();
    }
}


void GetAtom( void )
{
  //Die aktuellen Werte speichern.
  strcpy( actualFile->atom.last, actualFile->atom.actual );
  strcpy( actualFile->atom.lastID, actualFile->atom.actualID );

  strcpy( actualFile->atom.actual, actualFile->atom.next );
  strcpy( actualFile->atom.actualID, actualFile->atom.nextID );

  //Das naechste Zeichen einlesen.
  actualFile->GetChar();
  strcpy( actualFile->atom.next, Buffer() );

  CopyToBuffer( actualFile->atom.actual );

  //Nach einem Zeilenumbruch die Zeilen- und Spaltennummer anpassen.
  if( !strcmp( actualFile->atom.actual,"\n" ) )
    {
      actualFile->cursor.y++;
      actualFile->cursor.x = 0;
    }

  //Die Spaltennummer anpassen.
  actualFile->cursor.x++;

  //Das Dateiende muss man extra abfragen.
  if( actualFile->Eof() )
    {
      strcpy( actualFile->atom.next, "\0" );
      strcpy( actualFile->atom.nextID, "EOF" );
      return;
    }

  //Einen Bezeichner fuer das Zeichen ermitteln.
  switch( actualFile->atom.next[0] )
    {
    case '\x20':
      strcpy( actualFile->atom.nextID,"SPACE" );
      return;

    case '\n':
      strcpy( actualFile->atom.nextID,"NEWLINE" );

      //Fuer die Fehlerausgabe die Offsetposition speichern.
      actualFile->SetErrorOffset();

      return;

    case '\t':
      strcpy( actualFile->atom.nextID,"TAB" );
      return;

    case ';':
      strcpy( actualFile->atom.nextID,"SEMICOLON" );
      return;

    case '#':
      strcpy( actualFile->atom.nextID,"HASH" );
      return;

    case '$':
      strcpy( actualFile->atom.nextID,"DOLLAR" );
      return;

    case '{':
      strcpy( actualFile->atom.nextID,"OPEN_BRACE" );
      return;

    case '}':
      strcpy( actualFile->atom.nextID,"CLOSE_BRACE" );
      return;

    case '"':
      strcpy( actualFile->atom.nextID,"QUOTE" );
      return;

    case '[':
      strcpy( actualFile->atom.nextID,"OPEN_BRACKET" );
      return;

    case ']':
      strcpy( actualFile->atom.nextID,"CLOSE_BRACKET" );
      return;

    case '(':
      strcpy( actualFile->atom.nextID,"OPEN_PAREN" );
      return;

    case ')':
      strcpy( actualFile->atom.nextID,"CLOSE_PAREN" );
      return;

    case '\\':
      strcpy( actualFile->atom.nextID,"BACKSLASH" );
      return;

    default:
      strcpy( actualFile->atom.nextID,"NORMAL" );
      return;
    }
}


CCommand *GetCommand( char *name )
{
  return actualFile->handler.GetCommand( name );
}


void GetSymbol( void )
{
  //Die aktuellen Werte speichern.
  strcpy( actualFile->symbol.last, actualFile->symbol.actual );
  strcpy( actualFile->symbol.lastID, actualFile->symbol.actualID );

  strcpy( actualFile->symbol.actual, actualFile->symbol.next );
  strcpy( actualFile->symbol.actualID, actualFile->symbol.nextID );

  //Das erste Zeichen des Symbols einlesen.
  GetAtom();
  strcpy( actualFile->symbol.next, actualFile->atom.actual );
  strcpy( actualFile->symbol.nextID, actualFile->atom.actualID );

  CopyToBuffer( actualFile->symbol.actual );

  //Eventuell muessen noch mehr Zeichen eingelesen werden.
  switch( actualFile->symbol.next[0] )
    {
    case '\\':

      //Schon einmal den Bezeichner festlegen.
      strcpy( actualFile->symbol.nextID, "BSLASH_SUBST" );

      //Das Zeichen nach dem Schraegstrich einlesen.
      GetAtom();

      switch( actualFile->atom.actual[0] )
	{
	case 'a':
	case 'b':
	case 'f':
	case 'n':
	case 'r':
	case 't':
	case 'v':

	  //Es wurde ein Steuerzeichen angegeben. Das Zeichen wird einfach
	  //angehaengt.
	  strcat( actualFile->symbol.next, actualFile->atom.actual );
	  return;

	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':

	  //Es wurde eine oktale Zahl angegeben. Es koennen noch maximal zwei
	  //oktale Ziffern folgen.

	  //Die erste Ziffer anhaengen.
	  strcat( actualFile->symbol.next, actualFile->atom.actual );

	  //Schauen, ob das zweite Zeichen eine oktale Ziffer ist.
	  if( IsOctalDigit( actualFile->atom.next[0] ) )
	    {
	      //Die Ziffer wirklich einlesen.
	      GetAtom();

	      //Die zweite Ziffer anhaengen.
	      strcat( actualFile->symbol.next, actualFile->atom.actual );
	
	      //Schauen, ob das dritte Zeichen eine oktale Ziffer ist.
	      if( IsOctalDigit( actualFile->atom.next[0] ) )
		{
		  //Die Ziffer wirklich einlesen.
		  GetAtom();

		  //Die dritte Ziffer anhaengen.
		  strcat( actualFile->symbol.next, actualFile->atom.actual );
		}
	    }
	  return;

	case 'x':

	  //Es wurde eine hexadezimale Zahl angegeben (oder auch nur 'x').

	  //Das 'x' schon mal anhaengen.
	  strcat( actualFile->symbol.next, actualFile->atom.actual );

	  //Wenn eine hexadezimale Ziffern angegeben wurden, diese einlesen.
	  while ( IsHexDigit( actualFile->atom.next[0] ) )
	    {
	      GetAtom();
	      strcat( actualFile->symbol.next, actualFile->atom.actual );
	    }
	  return;

	case '\n':
	  //Es wurde ein Zeilenumbruch angegeben. Dies bedeutet, dass alle folgenden
	  //Leerzeichen zu einem einzigen Leerzeichen zusammengefasst werden.

	  //Alle Leerzeichen auslesen.
	  while( IsElementOf( actualFile->atom.nextID, " SPACE TAB " ) )
	    GetAtom();

	  strcpy( actualFile->symbol.next, " " );
	  strcpy( actualFile->symbol.nextID, "SPACE" );
	  return;

	default:
	  //Im Normalfall wird einfach das naechste Zeichen angehaengt.
	  strcat( actualFile->symbol.next, actualFile->atom.actual );
	  return;
	}

    case '(':
      if( IsGroupStart() )
	{
	  actualFile->stack.Push( parens, "OPEN" );
	  strcpy( actualFile->symbol.nextID, "START_PAREN_GROUP" );
	}
      return;

    case ')':
      if( IsGroupEnd() )
	{
	  actualFile->stack.Pop( parens );
	  strcpy( actualFile->symbol.nextID, "END_PAREN_GROUP" );
	}
      return;

    case '{':
      if( IsGroupStart() )
	{
	  actualFile->stack.Push( braces, "OPEN" );
	  strcpy( actualFile->symbol.nextID, "START_BRACE_GROUP" );
	}
      else
	{
	  //Eventuell befindet sich die Programmausfuehrung gerade in einem
	  //geklammerten Ausdruck.
	  if( actualFile->braceFlag )
	    actualFile->stack.Push( braces, "NORMAL" );
	}
      return;

    case '}':
      if( IsGroupEnd() )
	{
	  actualFile->stack.Pop( braces );
	  strcpy( actualFile->symbol.nextID, "END_BRACE_GROUP" );
	}
      else
	{
	  //Eventuell befindet sich die Programmausfuehrung gerade in einem
	  //geklammerten Ausdruck.
	  if( actualFile->braceFlag )
	    actualFile->stack.Pop( braces );
	}
      return;

    case '"':
      if( IsGroupStart() )
	{
	  actualFile->stack.Push( quotes, "OPEN" );
	  strcpy( actualFile->symbol.nextID, "START_QUOTE_GROUP" );
	}
      else if( IsGroupEnd() )
	{
	  actualFile->stack.Pop( quotes );
	  strcpy( actualFile->symbol.nextID, "END_QUOTE_GROUP" );
	}
      else
	strcpy( actualFile->symbol.nextID, "QUOTE" );

      return;

    case '[':
      if( IsGroupStart() )
	{
	  actualFile->stack.Push( brackets, "OPEN" );
	  strcpy( actualFile->symbol.nextID, "START_BRACKET_GROUP" );
	}
      return;

    case ']':
      if( IsGroupEnd() )
	{
	  actualFile->stack.Pop( brackets );
	  strcpy( actualFile->symbol.nextID, "END_BRACKET_GROUP" );
	}
      return;

    default:
      return;
    }
}


CVariable *GetVariable( char *name, char *environment )
{
  return actualFile->handler.GetVariable( name, environment );
}


void GetWord( void )
{
  //Die aktuellen Werte speichern.
  strcpy( actualFile->word.lastID, actualFile->word.actualID );
  strcpy( actualFile->word.last, actualFile->word.actual );

  //Vorherige Leerzeichen und Zeilenumbrueche auslesen.
  while( IsElementOf( actualFile->symbol.nextID, " SPACE TAB NEWLINE " ) )
    GetSymbol();

  //Um die Anfangsposition bei einem Fehler zu kennen, speichert man hier die
  //aktuelle Position als Fehlerposition ab.
  actualFile->error.x = actualFile->cursor.x;
  actualFile->error.y = actualFile->cursor.y;

  GetSymbol();

  //Ist das Zeichen eines mit einer besonderen Bedeutung, so wird es
  //zurueckgeliefert.
  if( IsElementOf( actualFile->symbol.actualID, " START_PAREN_GROUP END_PAREN_GROUP START_BRACE_GROUP END_BRACE_GROUP START_QUOTE_GROUP END_QUOTE_GROUP START_BRACKET_GROUP END_BRACKET_GROUP SEMICOLON HASH EOF " ) )
    {
      //Die Prozedur kann verlassen werden, da 'actual' und 'actualID' oben schon
      //angepasst wurden.
      strcpy( actualFile->word.actual, actualFile->symbol.actual );
      strcpy( actualFile->word.actualID, actualFile->symbol.actualID );
      CopyToBuffer( actualFile->word.actual );
      return;
    }      

  //Sonst werden die Zeichen in einer Schleife einzeln eingelesen.
  else
    {
      strcpy( actualFile->word.actual, "" );
      strcpy( actualFile->word.actualID, "" );

      //Innerhalb des Wortes werden Variablen- und Befehlssubstitutionen zugelassen.
      while( 1 )
	{
	  //Eine Befehlssubstitution behandeln.
	  if( !strcmp( actualFile->symbol.actualID, "START_BRACKET_GROUP" ) )
	    {
	      //Eventuellen Fehler abfangen.
	      if( !strcmp( ParseBracketGroup(), "" ) )
		{
		  strcpy( actualFile->word.actual, "" );
		  strcpy( actualFile->word.actualID, "ERROR" );
		  CopyToBuffer( "\0" );
		  return;
		}

	      strcat( actualFile->word.actual, Buffer() );
	    }

	  //Eine Variablensubstitution behandeln.
	  else if( !strcmp( actualFile->symbol.actualID, "DOLLAR" ) )
	    {
	      //Eventuellen Fehler abfangen.
	      if( !strcmp( ParseDollar(), "" ) )
		{
		  strcpy( actualFile->word.actual, "" );
		  strcpy( actualFile->word.actualID, "ERROR" );
		  CopyToBuffer( "\0" );
		  return;
		}

	      strcat( actualFile->word.actual, Buffer() );
	    }
	  else
	    strcat( actualFile->word.actual, actualFile->symbol.actual );

	  if( IsWordEnd() )
	    break;

	  //Das naechste Symbol einlesen.
	  GetSymbol();
	}
    }

  //Einen Bezeichner fuer das Wort bestimmen.
  if( IsProcedure( actualFile->word.actual ) )
    strcpy( actualFile->word.actualID, "PROCEDURE" );

  else if( IsTclCommand( actualFile->word.actual ) )
    strcpy( actualFile->word.actualID, "TCL_COMMAND" );

  else if( IsTkCommand( actualFile->word.actual ) )
    strcpy( actualFile->word.actualID, "TK_COMMAND" );

  else if( IsWidget( actualFile->word.actual ) )
    strcpy( actualFile->word.actualID, "WIDGET" );
    
  else if( IsImage( actualFile->word.actual ) )
    strcpy( actualFile->word.actualID, "IMAGE" );

  else
    strcpy( actualFile->word.actualID, "NORMAL" );

  CopyToBuffer( actualFile->word.actual );
}


void GetUnit( void )
{
  char dummy[ bufferSize ];

  //Die aktuellen Werte speichern.
  strcpy( actualFile->unit.lastID, actualFile->unit.actualID );
  strcpy( actualFile->unit.last, actualFile->unit.actual );

  //Das erste Wort der Einheit einlesen.
  GetWord();

  strcpy( actualFile->unit.actual, actualFile->word.actual );
  strcpy( actualFile->unit.actualID, actualFile->word.actualID );

  //Es liegt der Anfang eines Ausdrucks in geschweiften Klammern oder in
  //Anfuehrungszeichen vor. Der Ausdruck wird eingelesen, danach wird ein
  //Trennzeichen erwartet.
  if( IsElementOf( actualFile->word.actualID , " START_BRACE_GROUP START_QUOTE_GROUP " ) )
    {

      //Den entsprechenden Ausdruck einlesen.
      if( !strcmp( actualFile->word.actualID, "START_BRACE_GROUP" ) )
	strcpy( dummy, ParseBraceGroup() );

      else
	strcpy( dummy, ParseQuoteGroup() );

      //Eventuellen Fehler abfangen.
      if( !strcmp( dummy, "" ) )
	{
	  strcpy( actualFile->unit.actual, "" );
	  strcpy( actualFile->unit.actualID, "ERROR" );
	  CopyToBuffer( "\0" );
	  return;
	}

      //Nach dem Gruppenende muss ein Trennzeichen kommen.
      //if( IsWordEnd() == 0 )
      //{
      //  if( !strcmp( actualFile->word.actualID, "START_BRACE_GROUP" ) )
      //    display.Error( "unerlaubtes Zeichen nach Ende des {...}-Ausdrucks" );
      //  else
      //    display.Error( "unerlaubtes Zeichen nach Ende des \"...\"-Ausdrucks" );
      //}

      //Die Prozedur wird noch nicht verlassen, da am Ende noch ein passender Bezeichner
      //gefunden werden muss.
      strcpy( actualFile->unit.actual, dummy );
    }

  else if( !strcmp( "START_BRACKET_GROUP", actualFile->word.actualID ) )
    {

      //Eventuellen Fehler abfangen.
      if( !strcmp( ParseBracketGroup(), "" ) )
	{
	  strcpy( actualFile->unit.actual, "" );
	  strcpy( actualFile->unit.actualID, "ERROR" );
	  CopyToBuffer( "\0" );
	  return;
	}

      strcpy( actualFile->unit.actual, Buffer() );
    }

  //Sonst wird das Wort zurueckgeliefert.
  else
    {
      //Die Prozedur kann verlassen werden, da 'actual' und 'actualID' oben schon
      //angepasst wurden.
      CopyToBuffer( actualFile->unit.actual );
      return;
    }

  //Sonst einen Bezeichner fuer die Einheit bestimmen.
  if( IsProcedure( actualFile->unit.actual ) )
    strcpy( actualFile->unit.actualID, "PROCEDURE" );

  else if( IsTclCommand( actualFile->unit.actual ) )
    strcpy( actualFile->unit.actualID, "TCL_COMMAND" );

  else if( IsTkCommand( actualFile->unit.actual ) )
    strcpy( actualFile->unit.actualID, "TK_COMMAND" );

  else if( IsWidget( actualFile->unit.actual ) )
    strcpy( actualFile->unit.actualID, "WIDGET" );
    
  else if( IsImage( actualFile->unit.actual ) )
    strcpy( actualFile->unit.actualID, "IMAGE" );

  else
    strcpy( actualFile->unit.actualID, "NORMAL" );

  //Das Ergebnis zurueckliefern.
  CopyToBuffer( actualFile->unit.actual );
}


void HandleError( void )
{
  actualFile->EnableScanFlag();

  while( !IsLastArg() )
    GetUnit();

  actualFile->DisableScanFlag();
}


/* Mit 'SaveInfo()' gespeicherte Informationen laden */
/* ACHTUNG: actualFile muss auf ein Dateiobjekt zeigen */
int LoadInfo( char *name )
{
  char fileName[ strlen( name )+1 ];
  strcpy( fileName, name );

  //Ein Dateiobjekt zum Lesen erzeugen und versuchen, die Datei zu oeffnen.
  ifstream fin( fileName );

  //Eventuellen Fehler abfangen.
  if( fin.fail() )
    {
      display.Message( "information file '", fileName, "' does not exist" );
      return 0;
    }

  char line[ bufferSize ];
  char *dummy;

  //Die erste Zeile dient der Identifikation.
  fin.getline( line, unitLength, '\n' );

  //Fehler: die geoeffnete Datei ist keine Informationsdatei.
  if( strcmp( line, commandInfoID ) )
    {
      display.Message( "file '", fileName, "' is not a valid information file" );
      fin.close();
      return 0;
    }
    
  CCommand command;
  CVariable variable;

  //In der ersten Schleife werden die Befehlsinformationen ausgelesen.
  while( 1 )
    {
      //Die naechste Zeile auslesen.
      fin.getline( line, unitLength, '\n' );

      //Pruefen, ob das Ende des Befehlsabschnitts erreicht wurde.
      if( !strcmp( line, variableInfoID ) )
	break;

      //Eventuelles Dateiende abfragen.
      if( fin.eof() )
	{
	  display.Message( "error in information file'", fileName, "'" );
	  fin.close();
	  return 0;
	}

      //Die eingelesene Zeile aufteilen.
      dummy = strtok( line,"\t" );
      command.SetName( dummy );

      dummy = strtok( NULL,"\t" );
      command.SetCommandID( dummy );

      dummy = strtok( NULL,"\t" );
      command.SetList( dummy );

      dummy = strtok( NULL,"\t" );
      command.SetMin( atoi( dummy ) );

      dummy = strtok( NULL,"\t" );
      command.SetMax( atoi( dummy ) );

      dummy = strtok( NULL,"\t" );
      command.SetCalls( atoi( dummy ) );

      //Die Prozedur in das Programm aufnehmen.
      AddCommand( command );
    }

  //Danach die Variableninformationen auslesen.
  while( 1 )
    {
      //Die naechste Zeile einlesen.
      fin.getline( line, unitLength, '\n' );

      //Eventuelles Dateiende abfragen.
      if( fin.eof() )
	break;

      //Die eingelesene Zeile aufteilen.
      dummy = strtok( line,"\t" );
      variable.SetName( dummy );

      dummy = strtok( NULL,"\t" );
      variable.SetEnvironment( dummy );

      dummy = strtok( NULL,"\t" );
      variable.SetStatus( atoi( dummy ) );

      dummy = strtok( NULL,"\t" );
      variable.SetCalls( atoi( dummy ) );

      //Die Variable in das Programm aufnehmen.
      AddVariable( variable );
    }

  //Die Informationsdatei schliessen.
  fin.close();

  display.Message( "data from information file '", fileName, "' loaded" );
}


int LoadVars( char *name )
{
  char fileName[ strlen( name )+1 ];
  strcpy( fileName, name );

  //Ein Dateiobjekt zum Lesen erzeugen und versuchen, die Datei zu oeffnen.
  ifstream fin( fileName );

  //Eventuellen Fehler abfangen.
  if( fin.fail() )
    {
      display.Message( "system variables database '", fileName, "' not found, continuing without..." );
      return 0;
    }

  char line[ bufferSize ];

  while( 1 )
    {
      //Die naechste Zeile auslesen.
      fin.getline( line, unitLength, '\n' );

      //Eventuelles Dateiende abfragen.
      if( fin.eof() )
	break;

      //Die Variable sichtbar machen.
      AddVariable( line, " " );
    }

  fin.close();
  display.Message( "system variables database '", fileName, "' loaded" );
  return 1;
}


CCommand *ParseArgumentList( void )
{
  //Vorzeitiges Befehlsende abfangen.
  if( IsLastArg() )
    {
      display.Error( "Argumentenliste erwartet" );
      return NULL;
    }

  //Speicherplatz fuer das Ergebnis reservieren.
  CCommand *procedure = new CCommand;

  //Es gibt drei Moeglichkeiten:
  //
  //1. Liste in geschweiften Klammern.
  //2. Liste in Anfuehrungszeichen.
  //3. Ein einzelnes Argument.

  //Das Flag fuer Ausdruecke in geschweiften Klammern setzen. Dies ist notwendig, da
  //dort zu jeder oeffnenden Klammer eine schliessende Klammer existieren muss. Um diesen
  //Umstand kuemmert sich 'symbol.Get()', dem dies mit dem Flag signalisiert wird.
  actualFile->EnableBraceFlag();

  char result[ unitLength ] = "";
  int min = 0;
  int max = 0;
  int count = 0;

  GetWord();

  if( !strcmp( "START_BRACE_GROUP", ActualID( word ) ) )
    {

      while( 1 )
	{
	  GetWord();

	  if( !strcmp( "START_BRACE_GROUP", ActualID( word ) ) )
	    {
	      GetUnit();

	      if( !strcmp( "", ActualID( unit ) ) )
		display.Error( "argument without a name not allowed" );

	      else
		{
		  strcat( result, Actual( unit ) );
		  strcat( result, " " );
		  max++;
		}
	      count = 0;

	      while( 1 )
		{
		  GetUnit();

		  if( !strcmp( "END_BRACE_GROUP", ActualID( unit ) ) )
		    break;

		  else
		    count++;
		}

	      if( count > 1 )
		display.Error( "too many arguments in initialization list" );
	    }

	  else if( !strcmp( "END_BRACE_GROUP", ActualID( word ) ) )
	    {
	      if( !strcmp( Last( word ), "args" ) )
		{
		  min--;
		  max = -1;
		}

	      break;
	    }

	  else
	    {
	      strcat( result, Actual( word ) );
	      strcat( result, " " );
	      min++;
	      max++;
	    }
	}
    }

  else if( !strcmp( "START_QUOTE_GROUP", ActualID( word ) ) )
    {

      cout << "argument lists in quotes not yet supported\n";
      while( !strcmp( "END_QUOTE_GROUP", ActualID( word ) ) )
	{
	  GetWord();
	}
    }

  else
    {
      strcpy( result, Actual( word ) );
      min = 1;
      max = 1;

      if( !strcmp( Actual( word ), "args" ) )
	{
	  min--;
	  max = -1;
	}
    }

  if( !strcmp( result, "" ) )
    strcpy( result, " " );

  procedure->SetList( result );
  procedure->SetCommandID( "PROCEDURE" );
  procedure->SetMin( min );
  procedure->SetMax( max );

  actualFile->DisableBraceFlag();

  return procedure;
}


char *ParseBraceGroup( void )
{
  char result[ bufferSize ] = "\0";

  //Einen Zaehler fuer oeffnende Klammern anlegen. Die schon eingelesene
  //wird gleich mitgezaehlt.
  int openCount = 1;

  //Das Flag fuer Ausdruecke in geschweiften Klammern setzen. Dies ist notwendig, da
  //dort zu jeder oeffnenden Klammer eine schliessende Klammer existieren muss. Um diesen
  //Umstand kuemmert sich 'symbol.Get()', dem dies mit dem Flag signalisiert wird.
  actualFile->EnableBraceFlag();

  //In der Gruppe werden keine Substitutionen durchgefuehrt. Zeichen wie Zeilenumbrueche
  //oder Semikola haben hier auch keine besondere Bedeutung. Um Zeilenverlaengerungen
  //mittels Schraegstrich zu ermoeglichen, wird mit 'symbol.Get()' gearbeitet.
  while( openCount > 0 )
    {

      //Das naechste Symbol einlesen.
      GetSymbol();

      //Schauen, welche Bedeutung das Zeichen besitzt.
      if( !strcmp( Actual( symbol ), "{" ) )
	{
	  //Den Zaehler erhoehen und die Klammer dem Ergebnis hinzufuegen.
	  openCount++;
	  strcat( result, Actual( symbol ) );
	}

      else if( !strcmp( Actual( symbol ), "}" ) )
	{
	  //Den Zaehler erniedrigen und die Klammer dem Ergebnis hinzufuegen.
	  openCount--;

	  //Alle Klammern bis auf die letzte anhaengen.
	  if ( openCount > 0 )
	    strcat( result, Actual( symbol ) );
	}

      else if( !strcmp( ActualID( symbol ), "EOF" ) )
	{
	  //Wurde das Dateiende eingelesen, so wird abgebrochen.
	  display.Error( "end of file encountered in {...}-expression" );
	  actualFile->DisableBraceFlag();
	  CopyToBuffer( "\0" );
	  return Buffer();
	}

      else
	strcat( result, Actual( symbol ) );
    }

  //Das Flag zuruecksetzen.
  actualFile->DisableBraceFlag();
  CopyToBuffer( result );
  return Buffer();
}


char *ParseBracketGroup( void )
{
  char result[ strlen( "[ command-substitution ]" )+1 ];
  strcpy( result, "[ command-substitution ]" );

  //Bei einem Scanvorgang wird der Inhalt nicht kontrolliert.
  if( actualFile->scanFlag  )
    {
      int openCount = 1;

      while( openCount > 0 )
	{
	  //Das naechste Symbol einlesen.
	  GetSymbol();

	  //Schauen, welche Bedeutung das Zeichen besitzt.
	  if( !strcmp( Actual( symbol ), "[" ) )
	    openCount++;

	  else if( !strcmp( Actual( symbol ), "]" ) )
	    openCount--;

	  else if( !strcmp( ActualID( symbol ), "EOF" ) )
	    {
	      //Wurde das Dateiende eingelesen, so wird abgebrochen.
	      display.Error( "end of file encountered in [...]-expression" );
	      CopyToBuffer( "\0" );
	      return Buffer();
	    }
	}

      CopyToBuffer( result );
      return Buffer();
    }

  else
    {
      //Sonst in einer Schleife den Inhalt abarbeiten.
      while( 1 )
	{
	  //Die naechste Einheit einlesen.
	  GetUnit();

	  //Pruefen, ob das Ende des Ausdrucks erreicht wurde.
	  if ( !strcmp( ActualID( unit ), "END_BRACKET_GROUP" ) )
	    {
	      CopyToBuffer( result );
	      return Buffer();
	    }

	  //Es liegt ein Tcl-Befehl vor.
	  else if( !strcmp( ActualID( unit ), "TCL_COMMAND" ) )
	    ParseTclCommand( Actual( unit ) );

	  //Es liegt ein Tk-Befehl vor.
	  else if( !strcmp( ActualID( unit ), "TK_COMMAND" ) )
	    ParseTkCommand( Actual( unit ) );

	  //Es liegt ein Befehl in Form einer neuen Prozedur vor.
	  else if( !strcmp( ActualID( unit ), "PROCEDURE" ) )
	    ParseProcedure( Actual( unit ) );

	  //Es liegt ein erzeugtes Widget vor.
	  else if( !strcmp( ActualID( unit ), "WIDGET" ) )
	    ParseTkWidget( Actual( unit ) );

	  else if( !strcmp( ActualID( unit ), "IMAGE" ) )
	    ParseTkImage( Actual( unit ) );

	  //Es liegt ein Kommentar vor.
	  else if( !strcmp( ActualID( unit ), "HASH" ) )
	    ParseComment();

	  //Es liegt das Dateiende vor.
	  else if( !strcmp( ActualID( unit ), "EOF" ) )
	    {
	      display.Error( "end of file encountered in [...]-expression" );
	      CopyToBuffer( "\0" );
	      return Buffer();
	    }

	  //Es liegt ein Semikolon vor.
	  else if( !strcmp( ActualID( unit ), "SEMICOLON" ) )
	    {
	    }

	  //Sonst liegt ein unbekannter Befehl vor.
	  else
	    ParseUnknownCommand( Actual( unit ) );
	}
    }
}


int ParseCgetOption( char *commandID )
{
  char ID[ strlen( commandID )+1 ];
  strcpy( ID, commandID );

  if( !ParseUnit( "option" ) )
    return 0;

  //Ermitteln, ob es sich um eine gueltige Option handelt.
  if( !IsOption( Actual( unit ), ID ) )
    display.Error( "unknown option '", Actual( unit ), "' for this command" );

  ParseCommandEnd();
  return 1;
}


int ParseCommandEnd( void )
{
  if( IsLastArg() )
    return 1;
  else
    {
      display.Error( "end of command expected" );
      HandleError();
      return 0;
    }
}


void ParseComment( void )
{
  //Es wird solange gelesen, bis man einen Zeilenumbruch einliest,
  //vor dem kein Schraegstrich steht.

  while( 1 )
    {
      //Die Bearbeitung nur bei einem echten Zeilenumbruch abbrechen, nicht bei
      //einer Zeilenverlaengerung mit dem Schraegstrich.
      if( !strcmp( NextID( atom ), "NEWLINE" ) && strcmp( ActualID( atom ), "BACKSLASH" ) )
	break;

      GetAtom();
    }

  GetAtom();

  //Die Werte fuer 'symbol' anpassen.
  strcpy( actualFile->symbol.next, "\n" );
  strcpy( actualFile->symbol.nextID, "NEWLINE" );
}


int ParseConfigureOptions( char *ID )
{
  char commandID[ strlen( ID )+1 ];
  strcpy( commandID, ID );

  //Ende ist erlaubt.
  if( !ParseOptionalUnit() )
    return 0;

  //Das Eingelesene sollte ein Schaltername sein.
  if( !IsOption( Actual( unit ), commandID ) )
    display.Error( "unknown option '", Actual( unit ), "' for this command" );

  //Ende ist erlaubt.
  if( !ParseOptionalUnit() )
    return 1;

  //Das eingelesene war ein Argument, jetzt koennen noch beliebig viele solcher
  //'Option'-'Argument'-Paare folgen.

  //Es wird generell nur ein Argument eingelesen, wobei ein geklammerter Ausdruck auch
  //ein einziges Argument ist.

  while( ParseOptionalUnit() )
    {
      //Ist das gelesene ein gueltiger Schaltername.
      if( !IsOption( Actual( unit ), commandID ) )
	display.Error( "unknown option '", Actual( unit ), "' for this command" );

      //Das zugehoerige Argument einlesen.
      if( !ParseUnit( "value" ) )
	return 0;
    }
  return 1;
}


char *ParseDollar( void )
{
  //Speicherplatz fuer das Ergebnis reservieren und es initialisieren.
  char result[ bufferSize ];
  strcpy( result, "$" );

  char dummy[ bufferSize ];

  //Diese Variablen werden eventuell benoetigt.
  char arrayName[ wordLength ] = "\0";
  char indexName[ wordLength ] = "\0";
  int subst = 0;

  //Schauen, ob das naechste Symbol eine oeffnende Klammer ist.
  if( !strcmp( NextID( symbol ), "START_BRACE_GROUP" ) )
    {
      //Alle Zeichen bis zur schliessenden Klammer einlesen.
      GetSymbol();

      if( !strcmp( ParseBraceGroup(), "" ) )
	{
	  CopyToBuffer( "\0" );
	  return Buffer();
	}

      strcat( result, Buffer() );
    }

  //Sonst eine Zeichenkette einlesen.
  else
    {
      //Die erlaubten Zeichen festlegen.
      strcpy( dummy, "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_" );

      //Solange einlesen, bis ein unerlaubtes Zeichen kommt.
      while( strstr( dummy, Next( symbol ) ) != NULL )
	{
	  GetSymbol();
	  strcat( result, Actual( symbol ) );
	}

      //Kam kein gueltiges Zeichen nach dem Dollarzeichen, so wird nur das
      //Dollarzeichen zurueckgeliefert.
      if( !strcmp( result, "$" ) )
	{
	  CopyToBuffer( result );
	  return Buffer();
	}

      //Ist das naechste Zeichen eine oeffnende runde Klammer, so handelt es
      //sich um ein Feldelement.
      if( !strcmp( Next( symbol ), "(" ) )
	{
	  //Den Feldnamen speichern.
	  strcpy( arrayName, &result[1] );

	  //Die Klammer einlesen und im Ergebnis speichern.
	  GetSymbol();
	  strcat( result, Actual( symbol ) );

	  //Es wird bis zur naechsten schliessenden Klammer gelesen. Dabei sind
	  //weitere Variablensubstitutionen sowie Befehlssubstitutionen zugelassen.
	  while( strcmp( Next( symbol ), ")" ) )
	    {
	      //Ein eventuelles Dateiende abfangen.
	      if( !strcmp( NextID( symbol ), "EOF" ) )
		{
		  display.Error( "end of file in (...)-expression encountered" );
		  CopyToBuffer( "\0" );
		  return Buffer();
		}

	      //Das naechste Zeichen einlesen.
	      GetSymbol();

	      //Eine Befehlssubstitution behandeln.
	      if( !strcmp( ActualID( symbol ), "START_BRACKET_GROUP" ) )
		{
		  //Eventuellen Fehler abfangen.
		  if( !strcmp( ParseBracketGroup(), "" ) )
		    {
		      CopyToBuffer( "\0" );
		      return Buffer();
		    }

		  strcat( result, Buffer() );
		  strcat( indexName, Buffer() );

		  //Es muss fuer eine spaetere Pruefung festgehalten werden, dass
		  //eine Substitution im Index stattfand.
		  subst = 1;
		}

	      //Eine Variablensubstitution behandeln.
	      else if( !strcmp( ActualID( symbol ), "DOLLAR" ) )
		{
		  //Eventuellen Fehler abfangen.
		  if( !strcmp( ParseDollar(), "" ) )
		    {
		      CopyToBuffer( "\0" );
		      return Buffer();
		    }

		  strcat( result, Buffer() );
		  strcat( indexName, Buffer() );

		  //Es muss fuer eine spaetere Pruefung festgehalten werden, dass
		  //eine Substitution im Index stattfand.
		  subst = 1;
		}

	      else
		{
		  //Das Zeichen anhaengen.
		  strcat( result, Actual( symbol ) );

		  //Der Indexname wird auch separat abgespeichert.
		  strcat( indexName, Actual( symbol ) );
		}
	    }

	  //Die schliessende Klammer noch auslesen und ans Ergebnis anhaengen.
	  GetSymbol();
	  strcat( result, Actual( symbol ) );
	}
    }

  CopyToBuffer( result );

  //Befindet sich das Programm in einer {...}-Gruppe oder in in einem Scanvorgang,
  //so wird hier abgebrochen.

  if( actualFile->scanFlag || actualFile->braceFlag )
    return Buffer();

  //Sonst wird noch geprueft, ob die Variable bereits definiert wurde.
  //Dazu benoetigt man den Variablennamen ohne das Dollarzeichen am Anfang.
  strcpy( dummy, &result[1] );

  //Gibt es keinen Indexnamen, so wird nur geprueft, ob die Variable existiert.
  if( !strcmp( indexName, "" ) )
    {
      //Handelt es sich um eine normale Variable?
      if( IsVariable( dummy ) )
	{
	  CopyToBuffer( result );
	  return Buffer();
	}

      //Handelt es sich um eine Feldvariable, so wird eine Fehlermelung ausgegeben.
      if( IsArray( dummy ) )
	display.Error( "variable '", dummy, "' is an array" );

      //Sonst handelt es sich um eine unbekannte Variable.
      else
	display.Error( "unknown variable: '", dummy, "'" );

      CopyToBuffer( result );
      return Buffer();
    }

  //Gibt es einen Indexnamen, so wird zuerst geprueft, ob das
  //Feld existiert.
  else
    {
      //Die Variable ist noch nicht initialisiert, also kann sie auch ein Feld sein.
      CVariable *variable = GetVariable( arrayName, Environment() );
      if( variable != NULL )
	{
	  if( variable->Status() )
	    display.Error( "variable already initialized as non-array" );

	  CopyToBuffer( result );
	  return Buffer();
	}

      //Fehler: Das Feld existiert nicht.
      if( !IsArray( arrayName ) )
	{
	  display.Error( "unknown array: '", arrayName, "'" );
	  CopyToBuffer( result );
	  return Buffer();
	}	      

      //Dann wird geprueft, ob der Index zulaessig ist, sofern keine
      //Substitutionen in ihm stattfanden.
      else
	{
	  if( subst == 0 )
	    {
	      if( !IsVariable( dummy ) )
		display.Error( "unknown index '", indexName, "' of array '", arrayName, "'" );

	      CopyToBuffer( result );
	      return Buffer();
	    }
	  //Finden im Index Substitutionen statt, so wuerde normalerweise
	  //die obige Fehlermeldung ausgegeben werden. Dies wird hierdurch verhindert.
	  else
	    {
	      CopyToBuffer( result );
	      return Buffer();
	    }
	}
    }
}


void ParseFile( char *fileName )
{
  CInputFile fin( fileName );
  if( !fin.Open() )
    return;

  actualFile = &fin;

  //Versuchen, die zugehoerige Informationsdatei zu laden.
  if( !strcmp( flag.infoFile, "" ) )
    {
      char infoFileName[ strlen( fileName )+5 ];
      strcpy( infoFileName, fileName );
      strcat( infoFileName, ".ini" );
      LoadInfo( infoFileName );
    }

  else
    LoadInfo( flag.infoFile );

  display.Title( "file '", fileName, "' is beeing parsed ( ", LongToChar( fin.Size() ), " bytes )" );

  //In Tk ist das Standardfenster immer aktiv.
  if( flag.tclmode == 0 )
    actualFile->handler.AddWidget( ".", "toplevel" );


  //Alle Tcl- und Tk-Variablen sichtbar machen.
  char *home = getenv( "HOME" );
  char fullName[ strlen( home ) + strlen( varsFileName ) + 2 ];

  strcpy( fullName, home );
  strcat( fullName, "/" );
  strcat( fullName, varsFileName );

  LoadVars( fullName );

  //Die Werte fuer 'atom' und 'symbol' initialisieren, da diese vorauslesen.
  GetAtom();
  GetSymbol();

  char command[ bufferSize ];

  while( 1 )
    {
      GetUnit();

      strcpy( command, Buffer() );

      if( !strcmp( "PROCEDURE", ActualID( unit ) ) )
	ParseProcedure( command );

      else if( !strcmp( "TCL_COMMAND", ActualID( unit ) ) )
	ParseTclCommand( command );

      else if( !strcmp( "TK_COMMAND", ActualID( unit ) ) )
	ParseTkCommand( command );

      else if( !strcmp( "WIDGET", ActualID( unit ) ) )
	ParseTkWidget( command );

      else if( !strcmp( "IMAGE", ActualID( unit ) ) )
	ParseTkImage( command );

      else if( !strcmp( "HASH", ActualID( unit ) ) )
	ParseComment();

      else if( !strcmp( "SEMICOLON", ActualID( unit ) ) )
	{
	}

      else if( !strcmp( "EOF", ActualID( unit ) ) )
	{
	  display.Title( "done parsing file '", fileName, "' ( ", LongToChar( fin.Size() ), " bytes )" );
	  fin.Close();
	  actualFile = NULL;
	  return;
	}

      else
	ParseUnknownCommand( command );
    }
}


// 'ParseGeneric' liest die Argumente eines Befehls ein und prueft, ob die Anzahl
// zulaessig ist. 'min' steht fuer die Mindestanzahl an zulaessigen Argumenten,
// 'max' fuer die Hoechstanzahl. Ist 'max' gleich '-1', so sind beliebig viele
// Argumente zulaessig, jedoch mindestens 'min'.
void ParseGeneric( int min, int max )
{
  //Einen Argumentenzaehler erzeugen.
  int count = 0;

  //Es werden alle Argumente eingelesen. Danach wird erst geprueft, ob
  //die Anzahl zulaessig ist.

  while( ParseOptionalUnit() )
    {
      //Zaehler anpassen.
      count++;

      //Um auf eventuell vergessene '$'-Zeichen zur Variablendereferenzierung
      //hinzuweisen, wird bei Variablennamen eine Warnung ausgegeben.
      if( IsVariable( Actual( unit ) ) )
	  display.Warning( "argument '", Actual( unit ), "' is a variable - eventually '$' is missing" );
    }

  //Jetzt wird geprueft, ob die Anzahl im angegebenen Bereich liegt.
  if( count < min )
    display.Error( "too few arguments" );

  else if( (count > max) && (max != -1) )
    display.Error( "too many arguments" );
}


/* Eine optionale Einheit einlesen */
int ParseOptionalUnit( void )
{
  if( IsLastArg() )
    return 0;
  else
    {
      GetUnit();
      return 1;
    }
}


/* Ein optionales Wort einlesen */
int ParseOptionalWord( void )
{
  if( IsLastArg() )
    return 0;
  else
    {
      GetWord();
      return 1;
    }
}


int ParseOptions( char *ID )
{
  char commandID[ strlen( ID )+1 ];
  strcpy( commandID, ID );

  int type = 0;

  //Zuerst die Art des Befehls ermitteln, um die weitere vorgehensweise
  //bestimmen zu koennen. Naeheres siehe 'const.h'.
  if( IsElementOf( commandID, optionType1 ) )
    type = 1;

  else if( IsElementOf( commandID, optionType2 ) )
    type = 2;

  else if( IsElementOf( commandID, optionType3 ) )
    type = 3;

  else
    {
      display.InternalError( "ParseOptions", commandID );
      HandleError();
      return 0;
    }

  int explicitEnd = 0;

  //Ermitteln, ob die Liste '--' enthaelt.
  if( IsOption( "--", commandID ) )
    explicitEnd = 1;

  int count = 0;
  int max = 0;

  char dummy[ bufferSize ];

  //Befehle, bei denen die Schalter optional angegeben werden koennen und bei
  //denen nach den Schaltern das Befehlsende folgt.
  if( type == 1 )
    {

      //In einer Schleife alle Argumente einlesen und bearbeiten.
      while( ParseOptionalUnit() )
	{

	  strcpy( dummy, Actual( unit ) );
	  //Den Schalternamen auf Gueltigkeit pruefen.
	  if( !IsOption( dummy, commandID ) )
	    {
	      if( dummy[0] == '$' )
		display.Warning( "unknown option '", dummy, "' for this command" );

	      else
		display.Error( "unknown option '", dummy, "' for this command" );

	      HandleError();
	      return 0;
	    }

	  //Die zugehoerige Anzahl an Argumenten einlesen.
	  max = actualFile->handler.GetOptionCount( dummy, commandID );
	  count = 0;

	  while( 1 )
	    {
	      if( count >= max )
		break;

	      if( !ParseUnit() )
		return 0;

	      count++;
	    }
	}
      return 1;
    }

  else
    {
      while( 1 )
	{
	  if( IsLastArg() )
	    {
	      //Bei Typ 2 ist das Befehlsende nicht erlaubt, bei Typ 3 schon.
	      if( type == 2 )
		{
		  display.Error( "early command end" );
		  return 0;
		}

	      else
		return 1;
	    }

	  //Schauen, ob das naechste Argument ein gueltiger Schaltername ist.
	  if( !IsOption( Next( unit ), commandID ) )
	    return 1;

	  else
	    GetUnit();

	  //Ein Ende durch '--' pruefen.
	  if( explicitEnd && !strcmp( Actual( unit ), "--" ) )
	    return 1;

	  //Die zugehoerige Anzahl an Argumenten einlesen.
	  max = actualFile->handler.GetOptionCount( Actual( unit ), commandID );
	  count = 0;

	  while( 1 )
	    {
	      if( count >= max )
		break;

	      if( !ParseUnit() )
		return 0;

	      count++;
	    }
	}
    }
}


char *ParseParenGroup( void )
{
  cout << "ParseParenGroup not finished yet" << '\n';
  CopyToBuffer( "\0" );
  return Buffer();
}


void ParseProcedure( char *name )
{
  char nameCopy[ strlen( name )+1 ];
  strcpy( nameCopy, name );

  if( !IsProcedure( nameCopy ) )
    {
      display.InternalError( "ParseProcedure", nameCopy );
      HandleError();
      return;
    }

  actualFile->stack.Push( commands, nameCopy );

  CCommand *procedure = GetCommand( nameCopy );

  int min = procedure->Min();
  int max = procedure->Max();

  ParseGeneric( min, max );

  actualFile->stack.Pop( commands );
}


char *ParseQuoteGroup( void )
{
  char result[ bufferSize ] = "\0";

  //In der Gruppe sind Befehlsubstitutionen erlaubt.
  //In einer Schleife bis zum schliessenden Anfuehrungszeichen lesen.
  while( 1 )
    {
      //Das naechste Wort einlesen.
      GetSymbol();

      if( !strcmp( ActualID( symbol ) ,"END_QUOTE_GROUP" ) )
	{
	  CopyToBuffer( result );
	  return Buffer();
	}

      //Es handelt sich um den Anfang einer Befehlssubstitution.
      else if( !strcmp( ActualID( symbol ), "START_BRACKET_GROUP" ) )
	{
	  if( !strcmp( ParseBracketGroup(), "" ) )
	    {
	      CopyToBuffer( "\0" );
	      return Buffer();
	    }

	  else
	    strcat( result, Buffer() );
	}

      //Es handelt sich um eine Variablensubstitution.
      else if( !strcmp( ActualID( symbol ), "DOLLAR" ) )
	{
	  if( !strcmp( ParseDollar(), "" ) )
	    {
	      CopyToBuffer( "\0" );
	      return Buffer();
	    }

	  else
	    strcat( result, Buffer() );
	}

      //Es wurde das Dateiende eingelesen -> Fehler.
      else if( !strcmp( ActualID( symbol ), "EOF" ) )
	{
	  //Wurde das Dateiende eingelesen, so wird abgebrochen.
	  display.Error( "end of file in \"...\"-expression encountered" );
	  CopyToBuffer( "\0" );
	  return Buffer();
	}

      //Sonst wird das Symbol an das Ergebnis angehaengt.
      else
	strcat( result, Actual( symbol ) );
    }
}


int ParseScript( void )
{
  if( IsLastArg() )
    {
      display.Error( "command body expected" );
      return 0;
    }

  char command[ bufferSize ];
  char commandID[ unitLength ];

  GetWord();

  if( !strcmp( "START_BRACE_GROUP", ActualID( word ) ) )
    {
      while( 1 )
	{
	  GetUnit();

	  strcpy( command, Actual( unit ) );
	  strcpy( commandID, ActualID( unit ) );

	  if( !strcmp( "PROCEDURE", commandID ) )
	    ParseProcedure( command );

	  else if( !strcmp( "TCL_COMMAND", commandID ) )
	    ParseTclCommand( command );

	  else if( !strcmp( "TK_COMMAND", commandID ) )
	    ParseTkCommand( command );

	  else if( !strcmp( "WIDGET", commandID ) )
	    ParseTkWidget( command );

	  else if( !strcmp( "IMAGE", commandID ) )
	    ParseTkImage( command );

	  else if( !strcmp( "HASH", commandID ) )
	    ParseComment();

	  else if( !strcmp( "SEMICOLON", commandID ) )
	    {
	    }

	  else if( !strcmp( "EOF", commandID ) )
	    {
	      display.Error( "end of file encountered in command body" );
	      return 0;
	    }

	  else if( !strcmp( "HASH", commandID ) )
	    ParseComment();

	  else if( !strcmp( "END_BRACE_GROUP", commandID ) )
	    return 1;

	  else
	    ParseUnknownCommand( command );
	}
    }

  else if( !strcmp( "START_QUOTE_GROUP", ActualID( word ) ) )
    {
      while( 1 )
	{
	  GetUnit();

	  strcpy( command, Actual( unit ) );
	  strcpy( commandID, ActualID( unit ) );

	  if( !strcmp( "PROCEDURE", commandID ) )
	    ParseProcedure( command );

	  else if( !strcmp( "TCL_COMMAND", commandID ) )
	    ParseTclCommand( command );

	  else if( !strcmp( "TK_COMMAND", commandID ) )
	    ParseTkCommand( command );

	  else if( !strcmp( "WIDGET", commandID ) )
	    ParseTkWidget( command );

	  else if( !strcmp( "IMAGE", commandID ) )
	    ParseTkImage( command );

	  else if( !strcmp( "HASH", commandID ) )
	    ParseComment();

	  else if( !strcmp( "SEMICOLON", commandID ) )
	    {
	    }

	  else if( !strcmp( "EOF", commandID ) )
	    {
	      display.Error( "end of file encountered in command body" );
	      return 0;
	    }

	  else if( !strcmp( "HASH", commandID ) )
	    ParseComment();

	  else if( !strcmp( "END_QUOTE_GROUP", commandID ) )
	    return 1;

	  else
	    ParseUnknownCommand( command );
	}
    }

  else if( !strcmp( "START_BRACKET_GROUP", ActualID( word ) ) )
    {
      if( !strcmp( ParseBracketGroup(), "" ) )
	{
	  HandleError();
	  return 0;
	}

      else
	{
	  ParseCommandEnd();
	  return 1;
	}
    }

  else if( !strcmp( "PROCEDURE", ActualID( word ) ) )
    {
      CCommand *procedure = GetCommand( Actual( word ) );
      if( (procedure->Min() != 0) || (procedure->Max() > 0) )
	{
	  display.Error( "command without arguments expected" );
	  HandleError();
	  return 1;
	}

      ParseCommandEnd();
      return 1;
    }

  else if( !strcmp( "TCL_COMMAND", ActualID( word ) ) )
    {
      if( !IsElementOf( Actual( word ), nullArgsTcl ) )
	{
	  display.Error( "command without arguments expected" );
	  HandleError();
	  return 1;
	}
	
      ParseCommandEnd();
      return 1;
    }

  else
    {
      display.Error( "unknown command instead of command body encountered: ", Actual( word ) );
      HandleError();
      return 1;
    }
}


void ParseTclCommand( char *command )
{
  actualFile->stack.Push( commands, command );

  display.TclCommand( command );

  if( !strcmp( command, "after" ) )
    ParseTclAfterCmd();

  else if( !strcmp( command, "append" ) )
    ParseTclAppendCmd();

  else if( !strcmp( command, "array" ) )
    ParseTclArrayCmd();

  else if( !strcmp( command, "break" ) )
    ParseTclBreakCmd();

  else if( !strcmp( command, "case" ) )
    ParseTclCaseCmd();

  else if( !strcmp( command, "catch" ) )
    ParseTclCatchCmd();

  else if( !strcmp( command, "cd" ) )
    ParseTclCdCmd();

  else if( !strcmp( command, "clock" ) )
    ParseTclClockCmd();

  else if( !strcmp( command, "close" ) )
    ParseTclCloseCmd();

  else if( !strcmp( command, "concat" ) )
    ParseTclConcatCmd();

  else if( !strcmp( command, "continue" ) )
    ParseTclContinueCmd();

  else if( !strcmp( command, "eof" ) )
    ParseTclEofCmd();

  else if( !strcmp( command, "error" ) )
    ParseTclErrorCmd();

  else if( !strcmp( command, "eval" ) )
    ParseTclEvalCmd();

  else if( !strcmp( command, "exec" ) )
    ParseTclExecCmd();

  else if( !strcmp( command, "exit" ) )
    ParseTclExitCmd();

  else if( !strcmp( command, "expr" ) )
    ParseTclExprCmd();

  else if( !strcmp( command, "fblocked" ) )
    ParseTclFblockedCmd();

  else if( !strcmp( command, "fconfigure" ) )
    ParseTclFconfigureCmd();

  else if( !strcmp( command, "file" ) )
    ParseTclFileCmd();

  else if( !strcmp( command, "fileevent" ) )
    ParseTclFileeventCmd();

  else if( !strcmp( command, "flush" ) )
  ParseTclFlushCmd();

  else if( !strcmp( command, "for" ) )
    ParseTclForCmd();

  else if( !strcmp( command, "foreach" ) )
    ParseTclForeachCmd();

  else if( !strcmp( command, "format" ) )
    ParseTclFormatCmd();

  else if( !strcmp( command, "gets" ) )
    ParseTclGetsCmd();

  else if( !strcmp( command, "glob" ) )
    ParseTclGlobCmd();

  else if( !strcmp( command, "global" ) )
    ParseTclGlobalCmd();

  else if( !strcmp( command,"history" ) )
    ParseTclHistoryCmd();

  else if( !strcmp( command,"if" ) )
    ParseTclIfCmd();

  else if( !strcmp( command,"incr" ) )
    ParseTclIncrCmd();

  else if( !strcmp( command,"info" ) )
    ParseTclInfoCmd();

  else if( !strcmp( command,"interp" ) )
    ParseTclInterpCmd();

  else if( !strcmp( command,"join" ) )
    ParseTclJoinCmd();

  else if( !strcmp( command,"lappend" ) )
    ParseTclLappendCmd();

  else if( !strcmp( command,"lindex" ) )
    ParseTclLindexCmd();

  else if( !strcmp( command,"linsert" ) )
    ParseTclLinsertCmd();

  else if( !strcmp( command,"list" ) )
    ParseTclListCmd();

  else if( !strcmp( command,"llength" ) )
    ParseTclLlengthCmd();

  else if( !strcmp( command,"load" ) )
    ParseTclLoadCmd();

  else if( !strcmp( command,"lrange" ) )
    ParseTclLrangeCmd();

  else if( !strcmp( command,"lreplace" ) )
    ParseTclLreplaceCmd();

  else if( !strcmp( command,"lsearch" ) )
    ParseTclLsearchCmd();

  else if( !strcmp( command,"lsort" ) )
    ParseTclLsortCmd();

  else if( !strcmp( command,"open" ) )
    ParseTclOpenCmd();

  else if( !strcmp( command,"package" ) )
    ParseTclPackageCmd();

  else if( !strcmp( command,"pid" ) )
    ParseTclPidCmd();

  else if( !strcmp( command,"proc" ) )
    ParseTclProcCmd();

  else if( !strcmp( command,"puts" ) )
    ParseTclPutsCmd();

  else if( !strcmp( command,"pwd" ) )
    ParseTclPwdCmd();

  else if( !strcmp( command,"read" ) )
    ParseTclReadCmd();

  else if( !strcmp( command,"regexp" ) )
    ParseTclRegexpCmd();

  else if( !strcmp( command,"regsub" ) )
    ParseTclRegsubCmd();

  else if( !strcmp( command,"rename" ) )
    ParseTclRenameCmd();

  else if( !strcmp( command,"return" ) )
    ParseTclReturnCmd();

  else if( !strcmp( command,"scan" ) )
    ParseTclScanCmd();

  else if( !strcmp( command,"seek" ) )
    ParseTclSeekCmd();

  else if( !strcmp( command,"set" ) )
    ParseTclSetCmd();

  else if( !strcmp( command,"socket" ) )
    ParseTclSocketCmd();

  else if( !strcmp( command,"source" ) )
    ParseTclSourceCmd();

  else if( !strcmp( command,"split" ) )
    ParseTclSplitCmd();

  else if( !strcmp( command,"string" ) )
    ParseTclStringCmd();

  else if( !strcmp( command,"subst" ) )
    ParseTclSubstCmd();

  else if( !strcmp( command,"switch" ) )
    ParseTclSwitchCmd();

  else if( !strcmp( command,"tell" ) )
    ParseTclTellCmd();

  else if( !strcmp( command,"time" ) )
    ParseTclTimeCmd();

  else if( !strcmp( command,"trace" ) )
    ParseTclTraceCmd();

  else if( !strcmp( command,"unknown" ) )
    ParseTclUnknownCmd();

  else if( !strcmp( command,"unset" ) )
    ParseTclUnsetCmd();

  else if( !strcmp( command,"unsupported0" ) )
    ParseTclUnsupported0Cmd();

  else if( !strcmp( command,"update" ) )
    ParseTclUpdateCmd();

  else if( !strcmp( command,"uplevel" ) )
    ParseTclUplevelCmd();

  else if( !strcmp( command,"upvar" ) )
    ParseTclUpvarCmd();

  else if( !strcmp( command,"vwait" ) )
    ParseTclVwaitCmd();

  else if( !strcmp( command,"while" ) )
    ParseTclWhileCmd();

  else
    display.InternalError( "ParseTclCommand", command );

  actualFile->stack.Pop( commands );
}


void ParseTkCommand( char *command )
{
  actualFile->stack.Push( commands, command );

  display.TkCommand( command );

  if( IsElementOf( command, " button canvas checkbutton entry frame label listbox menu menubutton message radiobutton scale scrollbar text toplevel " ) )
    ParseTkWidgetCmd( command );

  else if( !strcmp( command, "bell" ) )
    ParseTkBellCmd();

  else if( !strcmp( command, "bind" ) )
    ParseTkBindCmd();

  else if( !strcmp( command, "bindtags" ) )
    ParseTkBindtagsCmd();

  else if( !strcmp( command, "clipboard" ) )
    ParseTkClipboardCmd();

  else if( !strcmp( command, "destroy" ) )
    ParseTkDestroyCmd();

  else if( !strcmp( command, "event" ) )
    ParseTkEventCmd();

  else if( !strcmp( command, "focus" ) )
    ParseTkFocusCmd();

  else if( !strcmp( command, "grab" ) )
    ParseTkGrabCmd();

  else if( !strcmp( command, "grid" ) )
    ParseTkGridCmd();

  else if( !strcmp( command, "image" ) )
    ParseTkImageCmd();

  else if( !strcmp( command, "lower" ) )
    ParseTkLowerCmd();

  else if( !strcmp( command, "option" ) )
    ParseTkOptionCmd();

  else if( !strcmp( command, "pack" ) )
    ParseTkPackCmd();

  else if( !strcmp( command, "place" ) )
    ParseTkPlaceCmd();

  else if( !strcmp( command, "raise" ) )
    ParseTkRaiseCmd();

  else if( !strcmp( command, "selection" ) )
    ParseTkSelectionCmd();

  else if( !strcmp( command, "send" ) )
    ParseTkSendCmd();

  else if( !strcmp( command, "tk" ) )
    ParseTkTkCmd();

  else if( !strcmp( command, "tkwait" ) )
    ParseTkTkwaitCmd();

  else if( !strcmp( command, "winfo" ) )
    ParseTkWinfoCmd();

  else if( !strcmp( command, "wm" ) )
    ParseTkWmCmd();

  else
    display.InternalError( "ParseTkCommand", command );

  actualFile->stack.Pop( commands );
}


void ParseTkWidget( char *name )
{
  char nameCopy[ strlen( name )+1 ];
  strcpy( nameCopy, name );

  if( !IsWidget( nameCopy ) )
    {
      display.InternalError( "ParseTkWidget", nameCopy );
      HandleError();
      return;
    }

  actualFile->stack.Push( commands, nameCopy );

  display.Widget( nameCopy );

  //Zuerst den Typ des Widgets ermitteln.
  CCommand *widget = GetCommand( nameCopy );
  char type[ strlen( widget->List() )+1 ];
  strcpy( type, Buffer() );

  if( !strcmp( type, "button" ) )
    ParseTkButtonWidget();

  else if( !strcmp( type, "canvas" ) )
    ParseTkCanvasWidget();

  else if( !strcmp( type, "checkbutton" ) )
    ParseTkCheckbuttonWidget();

  else if( !strcmp( type, "entry" ) )
    ParseTkEntryWidget();

  else if( !strcmp( type, "frame" ) )
    ParseTkFrameWidget();

  else if( !strcmp( type, "label" ) )
    ParseTkLabelWidget();

  else if( !strcmp( type, "listbox" ) )
    ParseTkListboxWidget();

  else if( !strcmp( type, "menu" ) )
    ParseTkMenuWidget();

  else if( !strcmp( type, "menubutton" ) )
    ParseTkMenubuttonWidget();

  else if( !strcmp( type, "message" ) )
    ParseTkMessageWidget();

  else if( !strcmp( type, "radiobutton" ) )
    ParseTkRadiobuttonWidget();

  else if( !strcmp( type, "scale" ) )
    ParseTkScaleWidget();

  else if( !strcmp( type, "scrollbar" ) )
    ParseTkScrollbarWidget();

  else if( !strcmp( type, "text" ) )
    ParseTkTextWidget();

  else if( !strcmp( type, "toplevel" ) )
    ParseTkToplevelWidget();

  else
    display.InternalError( "ParseTkWidget", type );

  actualFile->stack.Pop( commands );
}


void ParseTkWidgetCmd( char *command )
{
  if( !ParseUnit( "pathName" ) )
    return;

  char name[ strlen( Actual( unit )+1 ) ];
  strcpy( name, Actual( unit ) );

  if( name[0] == '$' )
    {
      display.Warning( "unknown widget: '", name, "'" );
      HandleError();
      return;
    }

  else if( name[0] != '.' )
    {
      display.Error( "not a valid name for a widget: '", name, "'" );
      HandleError();
      return;
    }

  char parent[ strlen( ParentWidget( name ) )+1 ];
  strcpy( parent, Buffer() );

  if( !IsWidget( parent ) )
    {
      display.Error( "parent widget '", parent, "' does not exist" );
      HandleError();
      return;
    }

  AddWidget( name, command );

  if( !ParseOptions( command ) )
    return;
}


int ParseUnit( void )
{
  if( IsLastArg() )
    {
      display.Error( "argument expected" );
      return 0;
    }
  else
    {
      //Die Einheit einlesen.
      GetUnit();
      return 1;
    }
}


int ParseUnit( char *name )
{
  if( IsLastArg() )
    {
      display.Error( "argument '", name, "' expected" );
      return 0;
    }
  else
    {
      //Die Einheit einlesen.
      GetUnit();
      return 1;
    }
}


void ParseUnknownCommand( char *name )
{
  if( name[0] == '#' )
    ParseComment();

  else
    {
      if( !strcmp( "[ command-substitution ]", name ) )
	display.Warning( "unable to build command" );

      else if( name[0] =='$' )
	display.Warning( "unknown command: '", name, "'" );

      else
	display.Error( "unknown command: '", name, "'" );

      HandleError();
    }
}


int ParseWidget( void )
{
  if( IsLastArg() )
    {
      display.Error( "widget expected" );
      return 0;
    }
  else
    {
      //Das mutmassliche Widget einlesen.
      GetUnit();

      if( Actual( unit )[0] == '$' )
	display.Warning( "unknown widget: '", Actual( unit ), "'" );

      else if( Actual( unit )[0] != '.' )
	display.Error( "not a valid name for a widget: '", Actual( unit ), "'" );

      else if( !IsWidget( Actual( unit ) ) )
	display.Error( "unknown widget: '", Actual( unit ), "'" );

      return 1;
    }
}


int ParseWord( void )
{
  if( IsLastArg() )
    {
      display.Error( "argument expected" );
      return 0;
    }
  else
    {
      //Das Wort einlesen.
      GetWord();
      return 1;
    }
}


int RemoveImage( char *name )
{
  if( IsImage( name ) )
    {
      RemoveCommand( name );
      return 1;
    }

  else
    return 0;
}


int RemoveProcedure( char *name )
{
  if( IsProcedure( name ) )
    {
      RemoveCommand( name );
      return 1;
    }

  else
    return 0;
}


int RemoveCommand( char *name )
{
  return actualFile->handler.RemoveCommand( name );
}


int RemoveVariable( CVariable &variable )
{
  char name[ strlen( variable.Name() )+1 ];
  strcpy( name, Buffer() );

  char environment[ strlen( variable.Environment() )+1 ];
  strcpy( environment, Buffer() );

  return actualFile->handler.RemoveVariable( name, environment );
}


int RemoveVariable( char *name, char *environment )
{
  return actualFile->handler.RemoveVariable( name, environment );
}


int RemoveWidget( char *name )
{
  char nameCopy[ strlen( name )+1 ];
  strcpy( nameCopy, name );

  if( IsWidget( nameCopy ) )
    {
      RemoveCommand( nameCopy );
      return 1;
    }

  else
    return 0;
}


void SaveInfo( char *fileName )
{
  if( actualFile->handler.Write( fileName ) )
    display.Title( "data stored in information file '", fileName, "'" );
}


int ScanFile( char *fileName )
{
  char dummy[ bufferSize ];

  CInputFile fin( fileName );
  if( !fin.Open() )
    return 0;

  actualFile = &fin;

  //Eine Startmeldung ausgeben.
  display.Title( "file '", fileName, "' is being scanned ( ", LongToChar( fin.Size() ), " bytes )" );

  CCommand *command;
  CVariable variable;

  //Durch dieses Flag signalisieren, das gescannt wird.
  actualFile->EnableScanFlag();

  //In einer Schleife die Datei bearbeiten.
  while( 1 )
    {
      //Das naechste Symbol einlesen.
      GetSymbol();

      //Das Dateiende abfragen.
      if( !strcmp( ActualID( symbol ), "EOF" ) )
	break;

      //Befehle werden nur nach Zeilenumbruechen, Semikola, oeffnenden
      //eckigen oder geschweiften Klammern oder nach Anfuehrungszeichen gesucht.
      if( IsElementOf( ActualID( symbol ), " NEWLINE SEMICOLON START_BRACKET_GROUP START_BRACE_GROUP " ) )
	{
	  //Den Befehl einlesen.
	  GetUnit();
	}
      else
	continue;

      //Das Dateiende abfragen.
      if( !strcmp( ActualID( unit ), "EOF" ) )
	break;

      //Die unterschiedlichen Befehle bearbeiten. Zuerst werden Befehle abgefragt,
      //auch innerhalb von Ruempfen beruecksichtigt werden.

      if( !strcmp( Actual( unit ), "proc" ) )
	{
	  //Den Prozedurnamen einlesen.
	  if( !ParseUnit() )
	    continue;

	  //Den Prozedurnamen speichern.
	  strcpy( dummy, Actual( unit ) );

	  //Die Argumentenliste einlesen.
	  command = ParseArgumentList();

	  if( command == NULL )
	    continue;

	  //Das Element anpassen.
	  command->SetName( dummy );
	  command->SetCalls( 0 );

	  //Die Prozedur fuer das Programm definieren.
	  AddCommand( *command );

	  //Eine Meldung ausgeben.
	  display.NewProcedure( dummy, command->List() );
	  delete command;
	  continue;
	}
      else if( !strcmp( Actual( unit ), "global" ) )
	{
	  if( IsLastArg() )
	    {
	      display.Error( "variable expected" );
	      continue;
	    }

	  while( ParseOptionalUnit() )
	    {
	      variable.SetName( Actual( unit ) );
	      variable.SetEnvironment( " " );

	      AddVariable( variable );

	      //Eine Meldung ausgeben.
	      display.NewVariable( Actual( unit ) );
	    }
	  continue;
	}
      else if( !strcmp( Actual( unit ), "source" ) )
	{
	  continue;
	}

      //Die folgenden Befehle werden nur ausserhalb von Klammern beruecksichtigt.
      if( IsGroupActive( " " ) )
	continue;

      //Die Befehle 'append', 'lappend' und 'set' koennen alle gleiche behandelt werden.
      if( IsElementOf( Actual( unit ), " append lappend set " ) )
	{
	  //Den Variablennamen einlesen.
	  if( !ParseUnit() )
	    continue;

	  //Den Variablennamen speichern.
	  strcpy( dummy, Actual( unit ) );

	  //Folgt noch ein Argument, so handelt es sich um eine Definition.
	  if( IsLastArg() )
	    continue;

	  //Die Variable fuer das Programm definieren.
	  variable.SetName( dummy );
	  variable.SetEnvironment( " " );

	  AddVariable( variable );

	  //Eine Meldung ausgeben.
	  display.NewVariable( dummy );
	  continue;
	}

      //Die Befehle 'catch' und 'gets' koennen beide zusammen behandelt werden.
      else if( IsElementOf( Actual( unit ), " catch gets " ) )
	{
	  //Zuerst das Befehlsskript (catch) oder das Datei-Handle (gets) einlesen.
	  if( !ParseUnit() )
	    continue;

	  //Danach den optionalen Variablennamen einlesen.
	  if( !ParseOptionalUnit() )
	    continue;

	  //Wurde der Name angegeben, so wird die Variable fuer das Programm definiert.
	  variable.SetName( Actual( unit ) );
	  variable.SetEnvironment( " " );
	  variable.SetStatus( 1 );
	  AddVariable( variable );

	  //Eine Meldung ausgeben.
	  display.NewVariable( Actual( unit ) );
	  continue;
	}
      else if( !strcmp( Actual( unit ), "if" ) )
	continue;

      else if( !strcmp( Actual( unit ), "file" ) )
	{
	  //Den Unterbefehl einlesen.
	  if( !ParseUnit() )
	    continue;

	  //Handelt es sich nicht um 'lstat' oder 'stat', so wird abgebrochen.
	  if( IsElementOf( Actual( unit ), " lstat stat " ) )
	    continue;

	  //Den Dateinamen einlesen.
	  if( !ParseUnit() )
	    continue;

	  //Zuletzt den Variablennamen einlesen. Dabei handelt es sich um ein
	  //Feld, bei dem mehrere Elemente gesetzt werden.
	  if( !ParseUnit() )
	    continue;

	  //Die Indexnamen in einem String speichern.
	  char indexList[] = "atime ctime dev gid ino mode ntime nlink size type uid";

	  //Den ersten Indexnamen aus dem String heraustrennen.
	  char *token = strtok( indexList," " );

	  //Alle Elemente des Feldes definieren.
	  while( token != NULL )
	    {
	      //Den vollstaendigen Variablennamen bilden.
	      strcpy( dummy, Actual( unit ) );
	      strcpy( dummy, "(" );
	      strcpy( dummy, token );
	      strcpy( dummy, ")" );

	      //Die Variable fuer das Programm definieren.
	      variable.SetName( dummy );
	      variable.SetEnvironment( " " );
	      variable.SetStatus( 1 );
	      if( AddVariable( variable ) )
		display.NewVariable( dummy );

	      //Den naechsten Indexnamen ermitteln.
	      token = strtok( NULL," " );
	    }
	  continue;
	}
      else if( !strcmp( Actual( unit ), "regexp" ) )
	{
	  if( !ParseOptions( "regexp" ) )
	    continue;

	  if( !ParseUnit( "exp" ) )
	    continue;

	  if( !ParseUnit( "string" ) )
	    continue;

	  //Alle Argumente die jetzt noch folgen werden als Variablen sichtbar gemacht.
	  while( ParseOptionalUnit() )
	    AddVariable( Actual( unit ), Environment(), 1, 0 );

	  continue;
	}

      else if( !strcmp( Actual( unit ), "regsub" ) )
	{
	  if( !ParseOptions( "regsub" ) )
	    continue;

	  if( !ParseUnit( "exp" ) )
	    continue;

	  if( !ParseUnit( "string" ) )
	    continue;

	  if( !ParseUnit( "subSpec" ) )
	    continue;

	  if( !ParseUnit( "varName" ) )
	    continue;

	  //Die Variable noch sichtbar machen.
	  AddVariable( Actual( unit ), Environment(), 1, 0 );
	  continue;
	}

      else if( !strcmp( Actual( unit ), "scan" ) )
	{
	  if( !ParseUnit( "string" ) )
	    continue;

	  if( !ParseUnit( "format" ) )
	    continue;

	  if( !ParseUnit( "varName" ) )
	    continue;

	  //Die Variable sichtbar machen.
	  AddVariable( Actual( unit ), Environment(), 1, 0 );

	  //Jetzt koennen noch optionale Variablennamen folgen.
	  while( ParseOptionalUnit() )
	    AddVariable( Actual( unit ), Environment(), 1, 0 );

	  continue;
	}

      else if( !strcmp( Actual( unit ), "upvar" ) )
	{
	  if( !ParseUnit() )
	    continue;

	  //Pruefen, ob das 'level'-Argument angegeben wurde.
	  if( (Actual( unit )[ 0 ] == '#') || (IsDigit( Actual( unit )[ 0 ] )) )
	    {
	      if( !IsLevel( Actual( unit ) ) )
		continue;

	      if( !ParseUnit( "otherVar" ) )
		continue;
	    }

	  if( !ParseUnit( "myVar" ) )
	    continue;

	  //Die Variable sichtbar machen.
	  AddVariable( Actual( unit ), Environment(), 0, 0 );

	  //Es koennen noch optionale 'otherVar'-'myVar'-Paare folgen.
	  while( 1 )
	    {
	      if( !ParseOptionalUnit() )
		continue;

	      if( !ParseUnit( "myVar" ) )
		continue;

	      //Die Variable sichtbar machen.
	      AddVariable( Actual( unit ), Environment(), 0, 0 );
	    }

	  continue;
	}

      else
	continue;
    }

  //Das Flag wieder zuruecksetzen.
  actualFile->DisableScanFlag();

  //Eine Meldung zum Schluss ausgeben.
  display.Title( "done scanning file '", fileName, "' ( ", LongToChar( fin.Size() ), " bytes )" );

  //Die gefundenen Daten in einer Informationsdatei speichern.
  char infoFileName[ wordLength ];
  strcpy( infoFileName, fileName );
  strcat( infoFileName, ".ini" );
  SaveInfo( infoFileName );

  //Die gescannte Datei schliessen.
  fin.Close();
  actualFile = NULL;
}













char *Environment( void )
{
  return actualFile->stack.Concat( procedures, " " );
}


//Nur fuer dezimale Zahlen
char *LongToChar( long number )
{
  char *result = new char[10];
  strcpy( result, "" );
  int pos = 0;

  int n = 0;
  long count = 10000000;

  long neu = number;
  if( neu < 0 )
    {
      neu = neu * -1;
      result[ pos ] = '-';
      pos++;
    }
    
  //Die Groessenordnung feststellen.
  while( 1 )
    {
      if( count > neu )
	count = count / 10;

      else
	break;
    }

  //Die einzelnen Ziffern in einen String kopieren.
  while( 1 )
    {
      if( neu < count )
	result[ pos ] = '0';

      else
	{
	  n = neu / count;
	  neu = neu % count;

	  //Umstaendlich?!?
	  switch( n )
	    {
	    case 0:
	      result[ pos ] = '0';
	      break;

	    case 1:
	      result[ pos ] = '1';
	      break;

	    case 2:
	      result[ pos ] = '2';
	      break;

	    case 3:
	      result[ pos ] = '3';
	      break;

	    case 4:
	      result[ pos ] = '4';
	      break;

	    case 5:
	      result[ pos ] = '5';
	      break;

	    case 6:
	      result[ pos ] = '6';
	      break;

	    case 7:
	      result[ pos ] = '7';
	      break;

	    case 8:
	      result[ pos ] = '8';
	      break;

	    case 9:
	      result[ pos ] = '9';
	      break;
	    }
	}

      pos++;

      if( count == 1 )
	{
	  result[ pos ] = '\0';
	  return result;
	}

      count = count / 10;
    }
}


char *ParentWidget( char *string )
{
  char copy[ strlen( string )+1 ];
  strcpy( copy, string );

  return actualFile->handler.ParentWidget( copy );
}


char *NextEnvironment( char *string )
{
  char copy[ strlen( string )+1 ];
  strcpy( copy, string );

  return actualFile->handler.NextEnvironment( copy );
}


int IsArray( char *name )
{
  char copy[ strlen( name )+1 ];
  strcpy( copy, name );

  return IsArray( copy, Environment() );
}


int IsArray( char *name, char *environment )
{
  return actualFile->handler.IsArray( name, environment );
}


int IsCommand( char *name )
{
  return actualFile->handler.IsCommand( name );
}


/* Ermitteln, ob 'ch' fuer eine dezimale Ziffer steht */
int IsDigit( char ch )
{
  if( strchr( "0123456789", ch ) == NULL )
    return 0;

  else
    return 1;
}


int IsElementOf( char *element, char *set )
{
  char dummy[ strlen( element )+3 ];

  strcpy( dummy, " " );
  strcat( dummy, element );
  strcat( dummy, " " );

  if( strstr( set, dummy ) == NULL )
    return 0;

  else
    return 1;
}


int IsGlobal( char *name )
{
  return actualFile->handler.IsGlobal( name );
}


/* Ermitteln, ob das Programm gerade innerhalb einer Gruppe arbeitet */
int IsGroupActive( char *ch )
{
  //Je nach angebenem Zeichen wird durch den Inhalt des entsprechenden Stacks
  //ermittelt, ob sich eine Gruppe aktiv ist.

  int a,b,c,d;

  switch( ch[0] )
    {
    case '(':
    case ')':
      if( actualFile->stack.Contains( parens, "OPEN" ) )
	return 1;
      else
	return 0;

    case '{':
    case '}':
      if( actualFile->stack.Contains( braces, "OPEN" ) )
	return 1;
      else
	return 0;

    case '"':
      if( actualFile->stack.Contains( quotes, "OPEN" ) )
	return 1;
      else
	return 0;

    case '[':
    case ']':
      if( actualFile->stack.Contains( brackets, "OPEN" ) )
	return 1;
      else
	return 0;

    case ' ':
      a = actualFile->stack.Contains( parens, "OPEN" );
      b = actualFile->stack.Contains( braces, "OPEN" );
      c = actualFile->stack.Contains( quotes, "OPEN" );
      d = actualFile->stack.Contains( brackets, "OPEN" );
      if( a || b || c || d )
	return 1;
      else
	return 0;

    default:
      display.InternalError( "GroupActive",ch );
      return 0;
    }
}


/* Ermitteln, ob ein Gruppenende vorliegt */
int IsGroupEnd( void )
{
  //Das aktuelle Zeichen muss ')', '}', '"' oder ']' sein.
  switch( Actual( atom )[0] )
    {
    case ')':
      if( !strcmp( actualFile->stack.Top( parens ),"OPEN" ) )
	return 1;
      else
	return 0;

    case '}':
      if( !strcmp( actualFile->stack.Top( braces ),"OPEN" ) )
	return 1;
      else
	return 0;

    case '"':
      if( !strcmp( actualFile->stack.Top( quotes ),"OPEN" ) )
	return 1;
      else
	return 0;

    case ']':
      if( !strcmp( actualFile->stack.Top( brackets ),"OPEN" ) )
	return 1;
      else
	return 0;

    default:
      display.InternalError( "GroupEnd", Actual( atom ) );
      return 0;
    }
}


int IsGroupStart( void )
{
  //Das aktuelle Symbol muss '(', '{', '"' oder '[' sein.
  //Es muss beachtet werden, dass die Prozedur immer von 'GetSymbol' mit
  //dem Zeichen 'symbol.next' aufgerufen wird.

  //Anfuehrungszeichen muessen gesondert behandelt werden.
  if( !strcmp( Next( symbol ), "\"" ) )
    {

      //Das vorherige Symbol entscheidet...
      if( IsElementOf( ActualID( symbol ), " SPACE TAB NEWLINE SEMICOLON START_PAREN_GROUP START_BRACE_GROUP START_BRACKET_GROUP " ) )
	{

	  //...aber nur, wenn noch keine "..."-Gruppe aktiv ist.
	  if( !strcmp( actualFile->stack.Top( quotes ), "OPEN" ) )
	    return 0;

	  else
	    return 1;
	}
      else
	return 0;
    }

  else if( !strcmp( Next( symbol ), "(" ) )
    {

      //Das vorherige Symbol entscheidet.
      if( IsElementOf( ActualID( symbol ), " SPACE TAB NEWLINE SEMICOLON START_PAREN_GROUP START_BRACE_GROUP START_QUOTE_GROUP START_BRACKET_GROUP " ) ) 
	return 1;

      else
	return 0;
    }
  else if( !strcmp( Next( symbol ), "{" ) )
    {

      //Das vorherige Symbol entscheidet.
      if( IsElementOf( ActualID( symbol ), " SPACE TAB NEWLINE SEMICOLON DOLLAR START_PAREN_GROUP START_BRACE_GROUP START_QUOTE_GROUP START_BRACKET_GROUP" ) )
	return 1;

      else
	return 0;
    }
  else if( !strcmp( Next( symbol ), "[" ) )
    return 1;

  else
    {
      display.InternalError( "GroupStart", Next( symbol ) );
      return 0;
    }
}


/* Ermitteln, ob 'ch' fuer eine oktale Ziffer steht */
int IsHexDigit( char ch )
{
  if( strchr( "0123456789abcdefABCDEF", ch ) == NULL )
    return 0;

  else
    return 1;
}


int IsImage( char *name )
{
  return actualFile->handler.IsImage( name );
}


/* Ermitteln, ob das letzte Argument eines Befehls eingelesen wurde */
int IsLastArg( void )
{
  //Leerzeichen werden entfernt.
  while( IsElementOf( NextID( symbol ), " SPACE TAB " ) )
    GetSymbol();

  //Jetzt entscheidet das naechste Symbol.
  if( IsElementOf( NextID( symbol ), " NEWLINE SEMICOLON END_BRACKET_GROUP END_BRACE_GROUP END_QUOTE_GROUP EOF " ) )
    return 1;

  else
    return 0;
}


// Ermittelt, ob 'string' eine gueltige Angabe fuer das 'level'-Argument
// beim 'uplevel' bzw. 'upvar' Befehl ist.
int IsLevel( char *string )
{
  int len = strlen( string );

  //Absolute Angabe mit '#'
  if( string[ 0 ] == '#' )
    {
      //Das Zeichen alleine ist nicht zulaessig.
      if( len == 1 )
	{
	  display.Error( "digits expected after '#'" );
	  return 0;
	}

      else
	{
	  //Pruefen, ob die folgenden Zeichen alles Ziffern sind.
	  for( int i=1 ; i < len ; i++ )
	    {
	      if( !IsDigit( string[ i ] ) )
		{
		  display.Error( "digits expected after '#'" );
		  return 0;
		}
	    }

	  return 1;
	}
    }

  //Relative Angabe durch Zahl
  else if( IsDigit( string[ 0 ] ) )
    {
      if( !IsNumber( string ) )
	{
	  display.Error( "number expected" );
	  return 0;
	}

      return 1;
    }
}


/* Ermitteln, ob gerade ein Schleifenbefehl bearbeitet wird */
int IsLoopActive( void )
{
  //Es kommen die Befehle 'for', 'foreach' und 'while' in Frage.
  //Man durchsucht den Stack auf diese Zeichenketten. Leider dreimal.
  if( actualFile->stack.Contains( commands, "for" ) )
    return 1;

  else if( actualFile->stack.Contains( commands, "foreach" ) )
    return 1;

  else if( actualFile->stack.Contains( commands, "while" ) )
    return 1;

  else
    return 0;
}


/* Ermitteln, ob 'ch' eine positive Zahl ist */
int IsNumber( char *ch )
{
  int len = strlen( ch );

  if( len == 0 )
    return 0;

  for( int i=0; i < len ; i++ )
    {
      if( !IsDigit( ch[ i ] ) )
	return 0;
    }

  return 1;
}


/* Ermitteln, ob 'ch' fuer eine hexadezimale Ziffer steht */
int IsOctalDigit( char ch )
{
  if( strchr( "01234567", ch ) == NULL )
    return 0;

  else
    return 1;
}


int IsOption( char *option, char *command )
{
  char optionCopy[ strlen( option )+1 ];
  strcpy( optionCopy, option );

  char commandCopy[ strlen( command )+1 ];
  strcpy( commandCopy, command );

  return actualFile->handler.IsOption( optionCopy, commandCopy );
}


/* Ermitteln, ob 'name' der Name einer Prozedur ist */
int IsProcedure( char *name )
{
  return actualFile->handler.IsProcedure( name );
}


/* Ermitteln, ob 'name' der Name eines Tcl-Befehls ist */
int IsTclCommand( char *name )
{
  //Untersuchen, ob " name " in der globalen Variablen tclCommandList enthalten ist.
  if( IsElementOf( name, tclCommandList ) )
    return 1;

  else
    return 0;
}


/* Ermitteln, ob 'name' der Name eines Tk-Befehls ist */
int IsTkCommand( char *name )
{
  //Untersuchen, ob " name " in der globalen Variablen tkCommandList enthalten ist.
  if( IsElementOf( name, tkCommandList ) )
    return 1;

  else
    return 0;
}


/* Ermitteln, ob 'name' der Name einer Variablen ist */
int IsVariable( char *name )
{
  char copy[ strlen( name )+1 ];
  strcpy( copy, name );

  return IsVariable( copy, Environment() );
}


/* Ermitteln, ob 'name' der Name einer Variablen ist */
int IsVariable( char *name, char *environment )
{
  return actualFile->handler.IsVariable( name, environment );
}


int IsWidget( char *name )
{
  return actualFile->handler.IsWidget( name );
}


/* Ermitteln, ob das letzte 'Symbol' eines Wortes eingelesen wurde */
int IsWordEnd( void )
{
  //Wie so oft entscheidet das naechste Symbol.
  if( IsElementOf( NextID( symbol ), " SPACE TAB NEWLINE SEMICOLON END_BRACKET_GROUP END_BRACE_GROUP END_QUOTE_GROUP END_PAREN_GROUP EOF " ) )
    return 1;

  else
    return 0;
}




