/*********************************************************************/
/* tclparse V2.00 from 14.March 1997                                 */
/* Copyright 1996/97 Stephan Weiss, Heiko Grossmann                  */
/*                   University of Stuttgart, Germany                */
/* For details see the copyright file included in the distribution.  */
/* This notice must not be removed from the file.                    */
/*                                                                   */
/* main source of the tclmerge program                               */
/*                                                                   */
/*********************************************************************/

#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include "../include/display.h"
#include "../include/parsing.h"
#include "../include/file.h"
#include "../include/handler.h"

int main( int argc, char *argv[] )
{
  //Eine Startmeldung ausgeben.
  display.Title( "TCLMERGE: melt several information files together for TCLPARSE" );

  //Fehler: Es wurde kein Dateiname angegeben.
  if( argc == 1 )
    {
      cout << "error: there were no filenames specified\n\n"
	   << "synopsis: merge  file1.tcl.ini  file2.tcl.ini  ? file3.tcl.ini ... ?  newName.ini\n\n";
      exit( 0 );
    }

  else if( argc == 2 )
    {
      cout << "error: for melting there are at least two files required\n\n"
	   << "synopsis: merge  file1.tcl.ini  file2.tcl.ini  ? file3.tcl.ini ... ?  newName.ini\n\n";
      exit( 0 );
    }

  else if( argc == 3 )
    {
      cout << "error: there was no name for the target file specified\n\n"
	   << "synopsis: merge  file1.tcl.ini  file2.tcl.ini  ? file3.tcl.ini ... ?  newName.ini\n\n";
      exit( 0 );
    }

  CInputFile file;
  actualFile = &file;

  //Ein Dateiobjekt zum Lesen erzeugen.
  ifstream fin;

  //Die Dateien nacheinander laden und zusammenfuegen.
  for( int i=1 ; i < argc-1 ; i++ )
    {
      fin.open( argv[ i ] );

      //Eventuellen Fehler abfangen.
      if( fin.fail() )
	{
	  display.Message( "information file '", argv[ i ], "' does not exist" );
	  return 0;
	}

      char *line = new char[ unitLength ];
      char *dummy = new char[ unitLength ];

      //Die erste Zeile dient der Identifikation.
      fin.getline( line, unitLength, '\n' );

      //Fehler: die geoeffnete Datei ist keine Informationsdatei.
      if( strcmp( line, commandInfoID ) )
	{
	  display.Message( "file '", argv[ i ], "' is not a valid information file" );
	  fin.close();
	  return 0;
	}
    
      CCommand command;
      CVariable variable;

      cout << "adding file '" << argv[ i ] << "' ...\n";

      //In der ersten Schleife werden die Befehlsinformationen ausgelesen.
      while( 1 )
	{
	  //Die naechste Zeile auslesen.
	  fin.getline( line, unitLength, '\n' );

	  //Pruefen, ob das Ende des Befehlsabschnitts erreicht wurde.
	  if( !strcmp( line, variableInfoID ) )
	    break;

	  //Eventuelles Dateiende abfragen.
	  if( fin.eof() )
	    {
	      display.Message( "error in information file '", argv[ i ], "'" );
	      fin.close();
	      return 0;
	    }

	  //Die eingelesene Zeile aufteilen.
	  dummy = strtok( line,"\t" );
	  command.SetName( dummy );

	  dummy = strtok( NULL,"\t" );
	  command.SetCommandID( dummy );

	  dummy = strtok( NULL,"\t" );
	  command.SetList( dummy );

	  dummy = strtok( NULL,"\t" );
	  command.SetMin( atoi( dummy ) );

	  dummy = strtok( NULL,"\t" );
	  command.SetMax( atoi( dummy ) );

	  dummy = strtok( NULL,"\t" );
	  command.SetCalls( atoi( dummy ) );

	  //Die Prozedur in das Programm aufnehmen.
	  AddCommand( command );
	}

      //Danach die Variableninformationen auslesen.
      while( 1 )
	{
	  //Die naechste Zeile einlesen.
	  fin.getline( line, unitLength, '\n' );

	  //Eventuelles Dateiende abfragen.
	  if( fin.eof() )
	    break;

	  //Die eingelesene Zeile aufteilen.
	  dummy = strtok( line,"\t" );
	  variable.SetName( dummy );

	  dummy = strtok( NULL,"\t" );
	  variable.SetEnvironment( dummy );

	  dummy = strtok( NULL,"\t" );
	  variable.SetStatus( atoi( dummy ) );

	  dummy = strtok( NULL,"\t" );
	  variable.SetCalls( atoi( dummy ) );

	  //Die Variable in das Programm aufnehmen.
	  AddVariable( variable );
	}

      //Die Informationsdatei schliessen.
      fin.close();
    }

  SaveInfo( argv[ argc-1 ] );
  return 1;
}














