/*********************************************************************/
/* tclparse V2.00 from 14.March 1997                                 */
/* Copyright 1996/97 Stephan Weiss, Heiko Grossmann                  */
/*                   University of Stuttgart, Germany                */
/* For details see the copyright file included in the distribution.  */
/* This notice must not be removed from the file.                    */
/*                                                                   */
/* routines for parsing tk commands                                  */
/*                                                                   */
/*********************************************************************/


#include "../include/tkcmd.h"
#include "../include/file.h"
#include "../include/parsing.h"
#include "../include/handler.h"
#include "../include/display.h"
#include <string.h>
#include <stdlib.h>


int ParseBindSequence( void )
{
  if( !ParseUnit( "sequence" ) )
    return 0;

  Actual( unit );

  int len = strlen( Buffer() );

  //Eine eventuelle Klammerung ueberpruefen.
  int startA = ( Buffer()[0] == '<' );
  int startB = ( Buffer()[1] == '<' );
  int endA = ( Buffer()[len-1] == '>' );
  int endB = ( Buffer()[len-2] == '>' );

  if( startA && startB )
    {
      if( !( endA && endB ) )
	display.Error( "expected '>>' at end of sequence" );
    }

  else if( startA )
    {
      if( !endA )
	display.Error( "expected '>' at end of sequence" );
    }

  return 1;
}


int ParseBindScript( void )
{

  if( !strcmp( Next( symbol ), "+" ) )
    {
      //Das '+'-Zeichen auslesen.
      GetSymbol();

      //Dann den Rumpf bearbeiten.
      if( !ParseScript() )
	return 0;
    }

  else
    {
      if( !ParseScript() )
	return 0;
    }

  return 1;
}


void ParseTkBellCmd( void )
{
  ParseOptions( "bell" );
  ParseCommandEnd();
}


void ParseTkBindCmd( void )
{
  if( !ParseUnit( "tag" ) )
    return;

  if( IsLastArg() )
    return;

  if( !ParseBindSequence() )
    return;

  if( IsLastArg() )
    return;

  if( !ParseBindScript() )
    return;

  ParseCommandEnd();
}


void ParseTkBindtagsCmd( void )
{
  if( !ParseWidget() )
    return;

  //Die optionalen 'tags' einlesen.
  while( ParseOptionalUnit() )
    {
    }
}


void ParseTkBitmapImage( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "image_bitmap" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "image_bitmap" );
      ParseCommandEnd();
    }
  
  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkButtonWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "button" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "button" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "flash" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "invoke" ) )
    ParseCommandEnd();

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkCanvasWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  char dummy[ bufferSize ];

  int count = 0;

  if( !strcmp( subCommand, "addtag" ) )
    {
      if( !ParseUnit( "tag" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "above" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "all" ) )
	ParseCommandEnd();

      else if( !strcmp( dummy, "below" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "closest" ) )
	ParseGeneric( 2, 4 );

      else if( !strcmp( dummy, "enclosed" ) )
	ParseGeneric( 4, 4 );

      else if( !strcmp( dummy, "overlapping" ) )
	ParseGeneric( 4, 4 );

      else if( !strcmp( dummy, "withtag" ) )
	ParseGeneric( 1, 1 );

      else
	{
	  display.Error( "unknown option: 'addtag ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "bbox" ) )
    ParseGeneric( 1, -1 );

  else if( !strcmp( subCommand, "bind" ) )
    ParseGeneric( 1, 3 );

  else if( IsElementOf( subCommand, " canvasx canvasy " ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "canvas" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "canvas" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "coords" ) )
    ParseGeneric( 1, -1 );

  else if( !strcmp( subCommand, "create" ) )
    {
      if( !ParseUnit( "type" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "arc" ) )
	{
	  //Es werden genau 4 Werte erwartet.
	  for( int i=1 ; i <= 4 ; i++ )
	    {
	      if( !ParseUnit( "value" ) )
		return;
	    }

	  if( !ParseOptions( "canvas_arc" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "bitmap" ) )
	{
	  //Es werden genau 2 Werte erwartet.
	  for( int i=1 ; i <= 2 ; i++ )
	    {
	      if( !ParseUnit( "value" ) )
		return;
	    }

	  if( !ParseOptions( "canvas_bitmap" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "image" ) )
	{
	  //Es werden genau 2 Werte erwartet.
	  for( int i=1 ; i <= 2 ; i++ )
	    {
	      if( !ParseUnit( "value" ) )
		return;
	    }

	  if( !ParseOptions( "canvas_image" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "line" ) )
	{
	  //Es werden mindestens 4 Werte erwartet, und immer Werte-Paare.
	  count = 0;
	  while( 1 )
	    {
	      //Das Ende der Werteparameter abfragen.
	      if( IsLastArg() || IsOption( Next( unit ), "canvas_line" ) )
		{
		  if( (count >= 4) && ( (count % 2) == 0) )
		    break;

		  else
		    {
		      display.Error( "wrong number of arguments" );
		      return;
		    }
		}

	      GetUnit();
	      count++;
	    }

	  if( !ParseOptions( "canvas_line" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "oval" ) )
	{
	  //Es werden genau 4 Werte erwartet.
	  for( int i=1 ; i <= 4 ; i++ )
	    {
	      if( !ParseUnit( "value" ) )
		return;
	    }

	  if( !ParseOptions( "canvas_oval" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "polygon" ) )
	{
	  //Es werden mindestens 6 Werte erwartet, und immer Werte-Paare.
	  count = 0;
	  while( 1 )
	    {
	      //Das Ende der Werteparameter abfragen.
	      if( IsLastArg() || IsOption( Next( unit ), "canvas_polygon" ) )
		{
		  if( (count >= 6) && ( (count % 2) == 0) )
		    break;

		  else
		    {
		      display.Error( "wrong number of arguments" );
		      return;
		    }
		}

	      GetUnit();
	      count++;
	    }

	  if( !ParseOptions( "canvas_polygon" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "rectangle" ) )
	{
	  //Es werden genau 4 Werte erwartet.
	  for( int i=1 ; i <= 4 ; i++ )
	    {
	      if( !ParseUnit( "value" ) )
		return;
	    }

	  if( !ParseOptions( "canvas_rectangle" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "text" ) )
	{
	  //Es werden genau 2 Werte erwartet.
	  for( int i=1 ; i <= 2 ; i++ )
	    {
	      if( !ParseUnit( "value" ) )
		return;
	    }

	  if( !ParseOptions( "canvas_text" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "window" ) )
	{
	  //Es werden genau 2 Werte erwartet.
	  for( int i=1 ; i <= 2 ; i++ )
	    {
	      if( !ParseUnit( "value" ) )
		return;
	    }

	  if( !ParseOptions( "canvas_text" ) )
	    return;

	  ParseCommandEnd();
	}

      else
	{
	  display.Error( "unknown option: 'create ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "dchars" ) )
    ParseGeneric( 2, 3 );

  else if( !strcmp( subCommand, "delete" ) )
    ParseGeneric( 0, -1 );

  else if( !strcmp( subCommand, "dtag" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "find" ) )
    ParseGeneric( 1, -1 );

  else if( !strcmp( subCommand, "focus" ) )
    ParseGeneric( 0, -1 );

  else if( !strcmp( subCommand, "gettags" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "icursor" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "index" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "insert" ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( subCommand, "itemcget" ) )
    {
      if( !ParseUnit( "tagOrId" ) )
	return;

      ParseCgetOption( "canvas_items" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "itemconfigure" ) )
    {
      if( !ParseUnit( "tagOrId" ) )
	return;

      ParseOptions( "canvas_items" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "lower" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "move" ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( subCommand, "postscript" ) )
    ParseOptions( "canvas_postscript" );

  else if( !strcmp( subCommand, "raise" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "scale" ) )
    ParseGeneric( 5, 5 );

  else if( !strcmp( subCommand, "scan" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "mark" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( dummy, "dragto" ) )
	ParseGeneric( 2, 2 );

      else
	{
	  display.Error( "unknown option: 'scan ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "select" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "adjust" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( dummy, "clear" ) )
	ParseCommandEnd();

      else if( !strcmp( dummy, "from" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( dummy, "item" ) )
	ParseCommandEnd();

      else if( !strcmp( dummy, "to" ) )
	ParseGeneric( 2, 2 );

      else
	{
	  display.Error( "unknown option: 'select ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "type" ) )
    ParseGeneric( 1, 1 );

  else if( IsElementOf( subCommand, " xview yview " ) )
    {
      if( !ParseOptionalUnit() )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "moveto" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "scroll" ) )
	ParseGeneric( 2, 2 );

      else
	{
	  display.Error( "unknown option: '", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
      return;
    }
}


void ParseTkCheckbuttonWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "checkbutton" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "checkbutton" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "deselect" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "flash" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "invoke" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "select" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "toggle" ) )
    ParseCommandEnd();

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkClipboardCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "clear" ) )
    {
      if( !ParseOptions( "clipboard_clear" ) )
	{
	  HandleError();
	  return;
	}

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "append" ) )
    {
      if( !ParseOptions( "clipboard_append" ) )
	{
	  HandleError();
	  return;
	}

      if( !ParseUnit( "data" ) )
	return;

      ParseCommandEnd();
    }

  else
    {
      display.UnknownOption( subCommand, "clipboard" );
      HandleError();
    }
}


void ParseTkDestroyCmd( void )
{
  while( ParseOptionalUnit() )
    {
      if( Actual( unit )[0] == '$' )
	{
	  display.Warning( "unknown widget: '", Actual( unit ) );
	  continue;
	}

      if( Actual( unit )[0] != '.' )
	{
	  display.Error( "not a valid name for a widget: '", Actual( unit ), "'" );
	  continue;
	}

      if( !RemoveWidget( Actual( unit ) ) )
	display.Error( "widget '", Actual( unit ), "' does not exist" );
    }
}


void ParseTkEntryWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  char dummy[ bufferSize ];

  if( !strcmp( subCommand, "bbox" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "cget" ) )
    ParseCgetOption( "entry" );

  else if( !strcmp( subCommand, "configure" ) )
    ParseConfigureOptions( "entry" );

  else if( !strcmp( subCommand, "delete" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "get" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "icursor" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "index" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "insert" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "scan" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "mark" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "dragto" ) )
	ParseGeneric( 1, 1 );

      else
	{
	  display.Error( "unknown option: 'scan ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "selection" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "adjust" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "clear" ) )
	ParseCommandEnd();

      else if( !strcmp( dummy, "from" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "present" ) )
	ParseCommandEnd();

      else if( !strcmp( dummy, "range" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( dummy, "to" ) )
	ParseGeneric( 1, 1 );

      else
	{
	  display.Error( "unknown option: 'selection ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "xview" ) )
    {
      if( !ParseOptionalUnit() )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "moveto" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "scroll" ) )
	ParseGeneric( 2, 2 );

      else
	ParseCommandEnd();
    }

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkEventCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "add" ) )
    {
      if( !ParseVirtualEvent() )
	return;

      if( !ParseBindSequence() )
	return;

      while( !IsLastArg() )
	{
	  if( !ParseBindSequence() )
	    return;
	}
    }

  else if( !strcmp( subCommand, "delete" ) )
    {
      if( !ParseVirtualEvent() )
	return;

      while( !IsLastArg() )
	{
	  if( !ParseBindSequence() )
	    return;
	}
    }

  else if( !strcmp( subCommand, "generate" ) )
    {
      if( !ParseWidget() )
	return;

      if( !ParseBindSequence() )
	return;

      if( !ParseOptions( "event generate" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "info" ) )
    {
      if( IsLastArg() )
	return;

      if( !ParseVirtualEvent() )
	return;

      ParseCommandEnd();
    }

  else
    {
      display.UnknownOption( subCommand, "event" );
      HandleError();
    }
}


void ParseTkFocusCmd( void )
{
  if( !ParseOptionalUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( IsElementOf( subCommand, " -displayof -force -lastfor " ) )
    {
      if( !ParseWidget() )
	return;

      ParseCommandEnd();
    }

  else
    {
      if( !IsWidget( Actual( unit ) ) )
	{
	  if( Actual( unit )[0] == '$' )
	    display.Warning( "unknown widget: '", Actual( unit ), "'" );

	  else
	    display.Error( "unknown widget: '", Actual( unit ), "'" );
	}

      ParseCommandEnd();
    }    
}


void ParseTkFrameWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "frame" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "frame" );
      ParseCommandEnd();
    }

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkGrabCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "-global" ) )
    {
      if( !ParseWidget() )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "current" ) )
    {
      if( IsLastArg() )
	return;

      if( !ParseWidget() )
	return;

      ParseCommandEnd();
    }

  else if( IsElementOf( subCommand, " release status " ) )
    {
      if( !ParseWidget() )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "set" ) )
    {
      if( !strcmp( Next( unit ), "-global" ) )
	GetUnit();

      if( !ParseWidget() )
	return;

      ParseCommandEnd();
    }

  else
    {
      display.UnknownOption( subCommand, "grab" );
      HandleError();
    }
}


void ParseTkGridCmd( void )
{
  //Ist ein Argument ein Widget, so wird dies nicht extra geprueft.

  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "bbox" ) )
    {
      if( !ParseUnit( "master" ) )
	return;

      if( IsLastArg() )
	return;

      if( !ParseUnit( "column" ) )
	return;

      if( !ParseUnit( "row" ) )
	return;
      
      if( IsLastArg() )
	return;

      if( !ParseUnit( "column2" ) )
	return;

      if( !ParseUnit( "row2" ) )
	return;

      ParseCommandEnd();
    }

  else if( IsElementOf( subCommand, " columnconfigure rowconfigure " ) )
    {
      if( !ParseUnit( "master" ) )
	return;

      if( !ParseUnit( "index" ) )
	return;

      if( !ParseOptions( "grid_rowconfigure" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      if( !ParseUnit( "slave" ) )
	return;

      //Bis zum Befehlsende oder vor den ersten Schalternamen lesen.
      while( !IsOption( Next( unit ), "grid" ) && !IsLastArg() )
	GetUnit();

      if( IsLastArg() )
	return;

      if( !ParseOptions( "grid" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "forget" ) )
    ParseGeneric( 1, -1 );

  else if( !strcmp( subCommand, "info" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "location" ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( subCommand, "propagate" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "remove" ) )
    ParseGeneric( 1, -1 );

  else if( !strcmp( subCommand, "size" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "slaves" ) )
    {
      if( !ParseUnit( "master" ) )
	return;

      if( !ParseOptions( "grid" ) )
	return;

      ParseCommandEnd();
    }

  else
    {
      //Bis zum Befehlsende oder vor den ersten Schalternamen lesen.
      while( !IsOption( Next( unit ), "grid" ) && !IsLastArg() )
	GetUnit();

      if( IsLastArg() )
	return;

      if( !ParseOptions( "grid" ) )
	return;

      ParseCommandEnd();
    }
}


void ParseTkImage( char *name )
{
  //Dieser Test ist eigentlich nicht notwendig, da die Prozedur immer mit
  //einem Bild aufgerufen wird. Aber man weiss ja nie.
  if( !IsImage( name ) )
    {
      display.InternalError( "ParseTkImage", name );
      HandleError();
      return;
    }

  //Den Typ des Bildes ermitteln.
  CCommand *image = GetCommand( name );
  if( !strcmp( image->List(), "BITMAP" ) )
    ParseTkBitmapImage();

  else if( !strcmp( image->List(), "PHOTO" ) )
    ParseTkPhotoImage();

  else
    {
      display.InternalError( "ParseTkImage", image->List() );
      HandleError();
    }
}


void ParseTkImageCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  char dummy[ bufferSize ];

  if( !strcmp( subCommand, "create" ) )
    {
      if( !ParseUnit( "type" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( IsLastArg() )
	return;

      if( !strcmp( dummy, "bitmap" ) )
	{
	  //Da der Name nicht angegeben werden muss, muss hier
	  //unterschieden werden.
	  if( !IsOption( Next( unit ), "image_bitmap" ) )
	    {
	      GetUnit();

	      //Das Bild als Befehl sichtbar machen.
	      AddImage( Actual( unit ), "BITMAP" );
	    }

	  if( !ParseOptions( "image_bitmap" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "photo" ) )
	{
	  if( !IsOption( Next( unit ), "image_photo" ) )
	    {
	      GetUnit();

	      //Das Bild als Befehl sichtbar machen.
	      AddImage( Actual( unit ), "PHOTO" );
	    }

	  if( !ParseOptions( "image_photo" ) )
	    return;

	  ParseCommandEnd();
	}
      
      else
	{
	  display.Error( "'bitmap' or 'photo' expected" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "delete" ) )
    ParseGeneric( 0, -1 );

  else if( IsElementOf( subCommand, " height type width " ) )
    ParseGeneric( 1, 1 );

  else if( IsElementOf( subCommand, " names types " ) )
    ParseCommandEnd();

  else
    {
      display.UnknownOption( subCommand, "image" );
      HandleError();
    }
}


void ParseTkLabelWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "label" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "label" );
      ParseCommandEnd();
    }

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkListboxWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  char dummy[ bufferSize ];

  if( !strcmp( subCommand, "activate" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "bbox" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "cget" ) )
    ParseCgetOption( "listbox" );

  else if( !strcmp( subCommand, "configure" ) )
    ParseConfigureOptions( "listbox" );

  else if( !strcmp( subCommand, "curselection" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "delete" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "get" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "index" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "insert" ) )
    ParseGeneric( 1, -1 );

  else if( !strcmp( subCommand, "nearest" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "scan" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "mark" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( dummy, "dragto" ) )
	ParseGeneric( 2, 2 );

      else
	{
	  display.Error( "unknown option: 'scan ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "see" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "selection" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "anchor" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "clear" ) )
	ParseGeneric( 1, 2 );

      else if( !strcmp( dummy, "includes" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "set" ) )
	ParseGeneric( 1, 2 );

      else
	{
	  display.Error( "unknown option: 'selection ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "size" ) )
    ParseCommandEnd();

  else if( IsElementOf( subCommand, " xview yview " ) )
    {
      if( !ParseOptionalUnit() )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "moveto" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "scroll" ) )
	ParseGeneric( 2, 2 );

      else
	ParseCommandEnd();
    }

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkLowerCmd( void )
{
  if( !ParseWidget() )
    return;

  ParseGeneric( 0, 1 );
}


void ParseTkMenuWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "activate" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "add" ) )
    {
      if( !ParseUnit( "type" ) )
	return;

      if( !IsElementOf( Actual( unit ), " cascade checkbutton command radiobutton separator " ) )
	{
	  display.Error( "not a valid option for argument 'type': '", Actual( unit ), "'" );
	  HandleError();
	  return;
	}

      if( !ParseOptions( "menu_add" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "menu" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "menu" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "delete" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "entrycget" ) )
    {
      if( !ParseUnit( "index" ) )
	return;

      ParseCgetOption( "menu" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "entryconfigure" ) )
    {
      if( !ParseUnit( "index" ) )
	return;

      ParseConfigureOptions( "menu" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "index" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "insert" ) )
    {
      if( !ParseUnit( "index" ) )
	return;

      if( !ParseUnit( "type" ) )
	return;

      if( !IsElementOf( Actual( unit ), " cascade checkbutton command radiobutton separator " ) )
	{
	  display.Error( "not a valid option for argument 'type': '", Actual( unit ), "'" );
	  HandleError();
	  return;
	}

      if( !ParseOptions( "menu add" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "invoke" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "post" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "postcascade" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "type" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "unpost" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "yposition" ) )
    ParseGeneric( 1, 1 );

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkMenubuttonWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "menubutton" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "menubutton" );
      ParseCommandEnd();
    }

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkMessageWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "message" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "message" );
      ParseCommandEnd();
    }

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkOptionCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "add" ) )
    ParseGeneric( 2, 3 );

  else if( !strcmp( subCommand, "clear" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "get" ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( subCommand, "readfile" ) )
    ParseGeneric( 1, 2 );

  else
    {
      display.UnknownOption( subCommand, "option" );
      HandleError();
    }
}


void ParseTkPackCmd( void )
{
  //Der Befehl besitzt eine alte und eine neue Syntax. Es werden
  //beide unterstuetzt.

  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  //neue Syntax.
  if( !strcmp( subCommand, "configure" ) )
    {
      if( !ParseWidget() )
	return;

      //Bis zum Befehlsende oder vor den ersten Schalternamen lesen.
      while( !IsOption( Next( unit ), "pack" ) && !IsLastArg() )
	ParseWidget();

      if( IsLastArg() )
	return;

      if( !ParseOptions( "pack" ) )
	return;

      ParseCommandEnd();
    }

  //neue Syntax.
  else if( !strcmp( subCommand, "forget" ) )
    {
      if( !ParseWidget() )
	return;

      while( !IsLastArg() )
	ParseWidget();

      ParseCommandEnd();
    }

  //neue Syntax.
  else if( IsElementOf( subCommand, " info slaves " ) )
    {
      if( !ParseWidget() )
	return;

      ParseCommandEnd();
    }

  //neue Syntax.
  else if( !strcmp( subCommand, "propagate" ) )
    {
      if( !ParseWidget() )
	return;

      ParseGeneric( 0, 1 );
    }

  //alte Syntax.
  else if( IsElementOf( subCommand, " after append before " ) )
    {
      //Das erste Widget einlesen.
      if( !ParseWidget() )
	return;

      //Das zweite Widget einlesen.
      if( !ParseWidget() )
	return;

      //Die Liste mit Optionen einlesen.
      if( !ParseUnit( "options" ) )
	return;

      //Jetzt koennen noch beliebig viele solcher Paare folgen.
      while( 1 )
	{
	  if( IsLastArg() )
	    return;

	  //Das Widget einlesen.
	  if( !ParseWidget() )
	    return;

	  //Die Liste mit Optionen muss aber bei angegebenem
	  //Widget folgen.
	  if( !ParseUnit( "options" ) )
	    return;
	}
    }

  //alte Syntax.
  else if( !strcmp( subCommand, "unpack" ) )
    {
      if( !ParseWidget() )
	return;

      ParseGeneric( 0, 1 );
    }

  else
    {
      //Bis zum Befehlsende oder vor den ersten Schalternamen lesen.
      while( !IsOption( Next( unit ), "pack" ) && !IsLastArg() )
	ParseWidget();

      if( IsLastArg() )
	return;

      if( !ParseOptions( "pack" ) )
	return;

      ParseCommandEnd();
    }
}


void ParseTkPhotoImage( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "blank" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "image_photo" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "image_photo" );
      ParseCommandEnd();
    }
  
  else if( !strcmp( subCommand, "copy" ) )
    {
      if( !ParseUnit( "sourceImage" ) )
	return;

      if( !ParseOptions( "image_photo_copy" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "get" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "put" ) )
    {
      if( !ParseUnit( "data" ) )
	return;

      if( !ParseOptions( "image_photo_put" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "read" ) )
    {
      if( !ParseUnit( "fileName" ) )
	return;

      if( !ParseOptions( "image_photo_read" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "redither" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "write" ) )
    {
      if( !ParseUnit( "fileName" ) )
	return;

      if( !ParseOptions( "image_photo_write" ) )
	return;

      ParseCommandEnd();
    }

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkPlaceCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "configure" ) )
    {
      if( !ParseWidget() )
	return;

      if( IsLastArg() )
	{
	  display.Error( "argument expected" );
	  return;
	}

      if( !ParseOptions( "pack" ) )
	return;

      ParseCommandEnd();
    }

  else if( IsElementOf( subCommand, " forget info slaves " ) )
    {
      if( !ParseWidget() )
	return;

      ParseCommandEnd();
    }

  else
    {
      if( !IsWidget( subCommand ) )
	{
	  if( subCommand[0] == '$' )
	    display.Warning( "unknown widget: '", subCommand, "'" );

	  else
	    display.Error( "unknown widget: '", subCommand, "'" );

	  HandleError();
	  return;
	}

      if( IsLastArg() )
	{
	  display.Error( "argument expected" );
	  return;
	}

      if( !ParseOptions( "place" ) )
	return;

      ParseCommandEnd();
    }
}


void ParseTkRadiobuttonWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "radiobutton" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "radiobutton" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "deselect" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "flash" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "invoke" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "select" ) )
    ParseCommandEnd();

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkRaiseCmd( void )
{
  if( !ParseWidget() )
    return;

  ParseGeneric( 0, 1 );
}


void ParseTkScaleWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "scale" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "scale" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "coords" ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( subCommand, "get" ) )
    {
      if( IsLastArg() )
	return;

      ParseGeneric( 2, 2 );
    }

  else if( !strcmp( subCommand, "identify" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "set" ) )
    ParseGeneric( 1, 1 );

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkScrollbarWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "scrollbar" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "scrollbar" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "delta" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "fraction" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "get" ) )
    ParseCommandEnd();

  else if( !strcmp( subCommand, "identify" ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( subCommand, "set" ) )
    ParseGeneric( 2, 2 );

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkSelectionCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "clear" ) )
    {
      if( !ParseOptions( "selection_clear" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "get" ) )
    {
      if( !ParseOptions( "selection_get" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "handle" ) )
    {
      if( !ParseOptions( "selection_handle" ) )
	return;

      ParseGeneric( 2, 2 );
      ParseCommandEnd();
    }

  //Nicht ganz korrekt was hier ablaeuft!
  else if( !strcmp( subCommand, "own" ) )
    {
      if( !ParseOptions( "selection_own" ) )
	return;

      ParseCommandEnd();
    }
}


void ParseTkSendCmd( void )
{
  if( !ParseOptions( "send" ) )
    return;

  ParseGeneric( 2, -1 );
}


void ParseTkTextWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  char dummy[ bufferSize ];

  if( !strcmp( subCommand, "bbox" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "cget" ) )
    {
      ParseCgetOption( "text" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "compare" ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( subCommand, "configure" ) )
    {
      ParseConfigureOptions( "text" );
      ParseCommandEnd();
    }

  else if( !strcmp( subCommand, "debug" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "dlineinfo" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "dump" ) )
    {
      if( !ParseOptions( "text_dump" ) )
	return;

      ParseGeneric( 1, 2 );
    }

  else if( !strcmp( subCommand, "get" ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( subCommand, "index" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "insert" ) )
    ParseGeneric( 2, -1 );

  else if( !strcmp( subCommand, "mark" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "gravity" ) )
	ParseGeneric( 1, 2 );

      else if( !strcmp( dummy, "names" ) )
	ParseCommandEnd();

      else if( !strcmp( dummy, "next" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "set" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( dummy, "unset" ) )
	ParseGeneric( 1, -1 );

      else
	{
	  display.Error( "unknown option: 'mark ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "scan" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "mark" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( dummy, "dragto" ) )
	ParseGeneric( 2, 2 );

      else
	{
	  display.Error( "unknown option: 'scan ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "search" ) )
    {
      if( !ParseOptions( "text_search" ) )
	return;

      ParseGeneric( 2, 3 );
    }

  else if( !strcmp( subCommand, "see" ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( subCommand, "tag" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );

      if( !strcmp( dummy, "add" ) )
	ParseGeneric( 2, -1 );

      else if( !strcmp( dummy, "bind" ) )
	{
	  if( !ParseUnit( "tagName" ) )
	    return;

	  if( IsLastArg() )
	    return;

	  //Ist das naechste Zeichen keine oeffnende Klammer, kein Anfuehrungszeichen und
	  //kein '+', so liegt das Argument 'sequence' vor.
	  if( !IsElementOf( Next( symbol ), " { \" + " ) )
	    ParseBindSequence();

	  if( IsLastArg() )
	    return;

	  if( !ParseBindScript() )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "cget" ) )
	{
	  if( !ParseUnit( "tagName" ) )
	    return;

	  if( !ParseCgetOption( "text_tag" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "configure" ) )
	{
	  if( !ParseUnit( "tagName" ) )
	    return;

	  if( !ParseConfigureOptions( "text_tag" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "delete" ) )
	ParseGeneric( 1, -1 );

      else if( !strcmp( dummy, "lower" ) )
	ParseGeneric( 1, 2 );

      else if( !strcmp( dummy, "names" ) )
	ParseGeneric( 0, 1 );

      else if( !strcmp( dummy, "nextrange" ) )
	ParseGeneric( 2, 3 );

      else if( !strcmp( dummy, "prevrange" ) )
	ParseGeneric( 2, 3 );

      else if( !strcmp( dummy, "raise" ) )
	ParseGeneric( 1, 2 );

      else if( !strcmp( dummy, "ranges" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "remove" ) )
	ParseGeneric( 2, -1 );

      else
	{
	  display.Error( "unknown option: 'tag ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "window" ) )
    {
      if( !ParseUnit( "option" ) )
	return;

      strcpy( dummy, Actual( unit ) );
	  
      if( !strcmp( dummy, "cget" ) )
	{
	  if( !ParseUnit( "index" ) )
	    return;

	  if( !ParseCgetOption( "text_window" ) )
	    return;

	  ParseCommandEnd();
	}

      else if( !strcmp( dummy, "configure" ) )
      {
	  if( !ParseUnit( "index" ) )
	    return;

	  if( !ParseConfigureOptions( "text_window" ) )
	    return;

	  ParseCommandEnd();
      }

      else if( !strcmp( dummy, "create" ) )
	{
	  if( !ParseUnit( "index" ) )
	    return;

	  if( !ParseConfigureOptions( "text_window" ) )
	    return;

	  ParseCommandEnd();
	}

      else
	{
	  display.Error( "unknown option: 'window ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "xview" ) )
    {
      if( !ParseOptionalUnit() )
	return;

      strcpy( dummy, Actual( unit ) );
	  
      if( !strcmp( dummy, "moveto" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "scroll" ) )
	ParseGeneric( 2, 2 );

      else
	{
	  display.Error( "unknown option: 'xview ", dummy, "'" );
	  HandleError();
	  return;
	}
    }

  else if( !strcmp( subCommand, "yview" ) )
    {
      if( !ParseOptionalUnit() )
	return;

      strcpy( dummy, Actual( unit ) );
	  
      if( !strcmp( dummy, "moveto" ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( dummy, "scroll" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( dummy, "-pickplace" ) )
	ParseGeneric( 1, 1 );

      else
	ParseCommandEnd();
    }

  else
    {
      display.Error( "unknown option: 'text ", subCommand, "'" );
      HandleError();
    }
}


void ParseTkTkCmd( void )
{
  if( !ParseUnit( "appname" ) )
    return;

  ParseGeneric( 0, 1 );
}


void ParseTkTkwaitCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "variable" ) )
    {
      if( !ParseUnit( "globalVar" ) )
	return;

      //Pruefen, ob das eingelesene eine globale Variable ist.
      if( !IsGlobal( Actual( unit ) ) )
	display.Error( "global variable expected instead of '", Actual( unit ), "'" );

      ParseCommandEnd();
    }

  else if( IsElementOf( subCommand, " visibility window " ) )
    {
      if( !ParseWidget() )
	return;

      ParseCommandEnd();
    }

  else
    {
      display.UnknownOption( subCommand, "tkwait" );
      HandleError();
    }
}


void ParseTkToplevelWidget( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( !strcmp( subCommand, "cget" ) )
    ParseCgetOption( "toplevel" );

  else if( !strcmp( subCommand, "configure" ) )
    ParseConfigureOptions( "toplevel" );

  else
    {
      display.Error( "unknown option: '", subCommand, "'" );
      HandleError();
    }
}


void ParseTkWinfoCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  if( IsElementOf( subCommand, " atom atomname containing interps pathname " ) )
    {

      if( !strcmp( Next( unit ), "-displayof" ) )
	{
	  GetUnit();

	  if( !ParseWidget() )
	    return;
	}

      if( IsElementOf( subCommand, " atom atomname pathname " ) )
	ParseGeneric( 1, 1 );

      else if( !strcmp( subCommand, "containing" ) )
	ParseGeneric( 2, 2 );

      else if( !strcmp( subCommand, "interps" ) )
	ParseCommandEnd();

      //Eigentlich ueberfluessig.
      else
	{
	  display.Error( "unknown command: 'winfo ", subCommand, "'" );
	  HandleError();
	}
    }

  //Alle anderen Befehle erwarten als zweites Argument ein Widget.
  else
    {
      if( !ParseWidget() )
	return;

      //Die folgenden Befehle erwarten nun das Befehlsende.
      if( IsElementOf( subCommand, " cells children class colormapfull depth exists geometry height id ismapped manager name parent pointerx pointerxy pointery reqheight reqwidth rootx rooty screen screencells screendepth screenheight screenmmheight screenmmwidth screenvisual screenwidth server toplevel viewable visual visualid vrootheight vrootwidth vrootx vrooty width x y " ) )
	ParseCommandEnd();

      else if( !strcmp( subCommand, "visualsavailable" ) )
	{
	  if( !ParseOptionalUnit() )
	    return;

	  //Sonst pruefen, ob 'includeids' angegeben wurde.
	  if( strcmp( Actual( unit ), "includeids" ) )
	    display.Error( "'includeids' erwartet anstatt '", Actual( unit ), "'" );

	  ParseCommandEnd();
	}

      else if( IsElementOf( subCommand, " fpixels pixels rgb " ) )
	ParseGeneric( 1, 1 );

      else
	{
	  display.UnknownOption( subCommand, "winfo" );
	  HandleError();
	}
    }
}


void ParseTkWmCmd( void )
{
  if( !ParseUnit() )
    return;

  char subCommand[ strlen( Actual( unit ) )+1 ];
  strcpy( subCommand, Buffer() );

  //Jeder Befehl benoetigt als zweites Argument den Namen eines Widgets. Deshalb
  //wird dieser Name hier EINMAL eingelesen.
  if( !ParseWidget() )
    return;

  //Jetzt den entsprechenden Unterbefehl nach dem Widgetnamen bearbeiten. 
  if( IsElementOf( subCommand, " aspect grid " ) )
    {
      if( IsLastArg() )
	return;

      else
	ParseGeneric( 4, 4 );
    }

  else if( !strcmp( subCommand, "client" ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( subCommand, "colormapwindows" ) )      
    ParseGeneric( 0, 1 );

  else if( !strcmp( subCommand, "command" ) )
    ParseGeneric( 0, 1 );

  else if( IsElementOf( subCommand, " deiconify frame iconify state withdraw " ) )      
    ParseCommandEnd();

  else if( !strcmp( subCommand, "focusmodel" ) )
    {
      if( IsLastArg() )
	return;

      else
	{
	  //Es wird 'active' oder 'passive' erwartet.
	  GetUnit();

	  if( !IsElementOf( Actual( unit ), " active passive " ) )
	    display.Error( "'active' or 'passive' expected" );

	  ParseCommandEnd();
	}
    }

  else if( !strcmp( subCommand, "geometry" ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( subCommand, "group" ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( subCommand, "iconbitmap" ) )      
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( subCommand, "iconmask" ) )      
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( subCommand, "iconname" ) )
    ParseGeneric( 0, 1 );

  else if( IsElementOf( subCommand, " iconposition maxsize minsize resizable " ) )      
    {
      if( IsLastArg() )
	return;

      else
	ParseGeneric( 2, 2 );
    }
    
  else if( !strcmp( subCommand, "iconwindows" ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( subCommand, "overrideredirect" ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( subCommand, "positionfrom" ) )      
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( subCommand, "protocol" ) )
    ParseGeneric( 0, 2 );

  else if( !strcmp( subCommand, "sizefrom" ) )      
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( subCommand, "title" ) )      
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( subCommand, "transient" ) )
    ParseGeneric( 0, 1 );

  else
    {
      display.UnknownOption( subCommand, "wm" );
      HandleError();
    }    
}


int ParseVirtualEvent( void )
{
  if( !ParseUnit( "<<virtual>>" ) )
    return 0;

  //Am Anfang muessen zwei oeffnende spitze Klammern stehen.
  if( (Actual( unit )[0] != '<') || (Actual( unit )[1] != '<') )
    {
      display.Error( "'<<' expected" );
      return 1;
    }

  //Am Ende muessen zwei schliessende spitze Klammern stehen.
  int len = strlen( Actual( unit ) );
  if( (Actual( unit )[ len-2 ] != '>') || (Actual( unit )[ len-1 ] != '>') )
    {
      display.Error( "'>>' expected" );
      return 1;
    }

  return 1;
}


